# NOTE: This package is copied directly from TaskManager.  It is
# being used as-is for now in the interest of expediency, despite
# its many deficiencies:
#
#   * Poor choice of package name
#   * Poor choice of method names
#   * Complete lack of file locking to protect against concurrent
#     access
#
# Eventually I would rather replace this with global eDAQ
# configuration (i.e. a registry approach) with SQLite3 on the back
# end, but there are space issues on the Classic that must be
# resolved before that can be seriously thought about. In the
# meantime, some of the more egregious deficiencies above could be
# fixed.
#
# --lareau

package SoMat::ConfigFile;

use Symbol ();
require SoMat::eDAQ::WebServer;

sub read_config_file {
    my ($s, $fname) = @_;
    my $fh = Symbol::gensym();
    return undef unless open($fh, "<$fname");
    local $/ = undef;
    my $content = <$fh>;
    return undef unless close($fh);
    return $content;
}

sub read_config_file_hash {
    my ($s, $fname, $h) = @_;
    my $content = $s->read_config_file($fname);
    return $h unless defined $content; 
    $h = {} unless defined $h;
    SoMat::eDAQ::WebServer->decode_query($content, $h);
    return $h;
}

sub write_config_file {
    my ($s, $fname, $contents, $mode) = @_;
    my $fh = Symbol::gensym();
    if (-f $fname) {
        my @st;
        my $len = length($contents);
        my $buf;
        my $ok = 1;
        $ok &&= (@st = stat($fname));
        my $stmode = $st[2] & 07777;
        $ok &&= (defined $mode ? $stmode == $mode : 1);
        $ok &&= ($st[7] == $len);
        $ok &&= open($fh, "<$fname");
        $ok &&= (read($fh, $buf, $len) == $len);
        $ok &&= close($fh);
        $ok &&= ($contents eq $buf);
        return 1 if $ok;
        unlink $fname;
    }
    (my $dir = $fname) =~ s/\/[^\/]+$//;
    unless (-d $dir) {
         my $dirmode = $mode;
         $dirmode = 0644 unless defined $dirmode;
         $dirmode |= 0100 if $dirmode & 0600;
         $dirmode |= 0010 if $dirmode & 0060;
         $dirmode |= 0001 if $dirmode & 0006;
         $dirmode = sprintf("0%o", $dirmode);
         system('/bin/mkdir', '-p', '-m', $dirmode, $dir) and return undef;
    }
    my $umask = (defined $mode) ? 0666 & ~$mode : undef;
    my $old_umask = umask $umask;
    my $ok = open($fh, ">$fname");
    umask $old_umask;
    return undef unless $ok;
    print $fh $contents;
    close $fh or return undef;
    return 1;
}

sub hash_to_config {
    my ($s, $h) = @_;
    my @cfg;
    while (my ($key, $val) = each %$h) {
        for my $v (@$val) {
            push(@cfg, "$key=" . SoMat::eDAQ::WebServer->escape_url($v));
        }
    }
    return join('&', @cfg);
}

sub write_config_file_hash {
    my ($s, $fname, $h, $mode) = @_;
    my $cfg = $s->hash_to_config($h);
    return $s->write_config_file($fname, $cfg, $mode);
}

sub read_pid_file {
    my ($s, $fname) = @_;
    my $pid = $s->read_config_file($fname);
    return undef unless defined $pid && $pid =~ /^\s*(\d+)\s*$/;
    return $1;
}

1;
