
package SoMat::File;
use strict;

use constant O_RDONLY    =>       00;
use constant O_WRONLY    =>       01;
use constant O_RDWR      =>       02;
use constant O_CREAT     =>     0100;
#use constant O_EXCL      =>     0200;
#use constant O_NOCTTY    =>     0400;
use constant O_TRUNC     =>    01000;
use constant O_APPEND    =>    02000;
#use constant O_NONBLOCK  =>    04000;
#use constant O_NDELAY    =>  O_NONBLOCK;
#use constant O_SYNC      =>   010000;
#use constant O_FSYNC     =>   O_SYNC;
#use constant O_ASYNC     =>   020000;

use Symbol ();

sub new {
    my $proto = shift;
    my $s = Symbol::gensym();
    bless $s, ref($proto) || $proto;
    return undef unless $s->open(@_);
    $s->set('DEBUG', 0);
    return $s;
}

sub DESTROY {
    my ($s) = @_;
    $s->close;
}

sub set {
    my ($s, $name, $value) = @_;
    ${*$s}{$name} = $value;
}

sub get {
    my ($s, $name) = @_;
    return ${*$s}{$name};
}

sub open {
    my ($s, $filename, $mode) = @_;
    ${*$s}{_filename} = $filename;
    ${*$s}{_filemode} = $mode;
    my $ret;
    if (defined $mode) {
        bless $s, 'SoMat::File::Sys'; # KLUDGE!
        $ret = CORE::sysopen($s, $filename, $mode);
    } else {
        bless $s, 'SoMat::File::Stdio'; # KLUDGE!
        $ret = CORE::open($s, $filename);
    }
    binmode($s) if $ret;
    ${*$s}{length} = -s $s;
    return $ret;
}

sub close {
    return CORE::close($_[0]);
}

sub contents { # Suck in entire file
    my ($s, $fname) = @_;
    my $fh = Symbol::gensym();
    return undef unless CORE::open($fh, "<$fname");
    local $/ = undef;
    my $content = <$fh>;
    return undef unless CORE::close($fh);
    return $content;
}

package SoMat::File::Temp;

my @Temps;
my $terminator;

sub _init {
    my ($s) = @_;
    return if defined $terminator;
    $terminator = bless {}, ref($s) || $s;
}

sub _tempname {
    my ($s, $dir, $pat) = @_;
    $pat =~ s/X+/$$ + int(rand(100))/e;
    return "$dir/$pat";
}

sub tempfile {
    my ($s, $pat, $dir) = @_;
    $s->_init;
    unless (defined $dir) {
        $dir = -d '/tmp' ? '/tmp' : $ENV{TMP} || $ENV{TEMP};
    }
    return undef unless -d $dir;
    $pat = 'tmpfileXXX' unless defined $pat;
    local *FH;

    my $tries = 100;
    while (--$tries) {
        my $fname = $s->_tempname($dir, $pat);
        # KLUDGE O_WRONLY = 01, O_EXCL = 0200, O_CREAT = 0100
        # if (sysopen(FH, $fname, O_WRONLY|O_EXCL|O_CREAT)) {
        if (sysopen(FH, $fname, 01 | 0200 | 0100)) {
            close(FH);
            push @Temps, $fname;
            return $fname;
        }
    }
    return undef;
}

sub tempdir {
    my ($s, $pat, $dir) = @_;
    $s->_init;
    $dir = '/tmp' unless defined $dir;
    return undef unless -d $dir;
    $pat = 'tmpdirXXX' unless defined $pat;
    my $tries = 10;
    while (--$tries) {
        my $temp = $s->_tempname($dir, $pat);
        if (mkdir($temp, 0777)) {
            push @Temps, $temp;
            return $temp
        }
    }
    return undef;
}

sub cleanup {
    my ($s, @files) = @_;
    @files = @Temps unless @files;
    for my $f (@files) {
        if (-d $f) {
            system('/bin/rm', '-rf', $f);
        } else {
            unlink $f; # Ignore errors; nothing we can do
        }
    }
}

sub DESTROY {
    cleanup();
}

package SoMat::File::Sys;

@SoMat::File::Sys::ISA = qw(SoMat::File);

sub read {
    return CORE::sysread($_[0], $_[1], $_[2], $_[3] || 0);
}

#sub read {
#    my $got = CORE::sysread($_[0], $_[1], $_[2], $_[3] || 0);
#    if ($got >= 0) {
#        my $s = $_[0];
#        my $fpos = ${*$s}{fpos};
#        print "READ $got bytes from $fpos\n" if $s->get('DEBUG');
#        my $data = $_[1];
#        $data =~ s/(.)/uc sprintf(" %02x", ord($1))/seg;
#        print "[$data ]\n" if $s->get('DEBUG');
#        ${*$s}{fpos} += $got;
#    }
#    return $got;
#}

sub write {
    my ($s, $data, $len, $off) = @_;
    $len = length($data) unless defined $len;
    $off = 0 unless defined $off;
    return CORE::syswrite($s, $data, $len, $off);
}

sub seek {
    return CORE::sysseek($_[0], $_[1], $_[2] || 0);
}

#sub seek {
#    my $fpos = CORE::sysseek($_[0], $_[1], $_[2] || 0);
#    my $s = $_[0];
#    ${*$s}{fpos} = $fpos;
#    print "SEEK: $fpos\n" if $s->get('DEBUG');
#    return $fpos;
#}

package SoMat::File::Stdio;

@SoMat::File::Stdio::ISA = qw(SoMat::File);

sub read {
    return CORE::read($_[0], $_[1], $_[2], $_[3] || 0);
}

sub write {
    my ($s, $data, $len, $off) = @_;
    unless (defined $len) {
        return length($data) if print $s $data;
        return undef;
    }
    $off = 0 unless defined $off;
    $data = substr($data, $off, $len);
    return length($data) if print $s $data;
    return undef;
}

sub seek {
    return undef unless CORE::seek($_[0], $_[1], $_[2] || 0);
    return tell($_[0]);
}

package SoMat::File::Offset;

@SoMat::File::Offset::ISA = qw(SoMat::File);

sub open {
    my ($s, $filename, $mode, $offset, $length) = @_;
    $mode = SoMat::File::O_RDONLY() unless defined $mode; # Force Sys
    my $ret = SoMat::File::open($s, $filename, $mode);
    bless $s, 'SoMat::File::Offset'; # KLUDGE!
    if ($ret) {
        ${*$s}{offset} = $offset;
        ${*$s}{length} = $length;
        ${*$s}{fpos} = 0;
        $s->seek(0, 0);
    }
    return $ret;
}

sub close {
    return CORE::close($_[0]);
}

sub read {
    my ($s, undef, $len, $off) = @_;
    $len = length($_[1]) unless defined $len;
    $off = 0 unless defined $off;
    return 0 if ${*$s}{fpos} >= ${*$s}{length}; 
    $len = ${*$s}{length} - ${*$s}{fpos}
        if ${*$s}{fpos} + $len > ${*$s}{length};
    my $ret = CORE::sysread($_[0], $_[1], $len, $off);
    ${*$s}{fpos} += $ret;
    return $ret;
}

sub write {
    return CORE::syswrite($_[0], $_[1], $_[2], $_[3]);
}

sub seek {
    my ($s, $pos, $whence) = @_;
    my $fpos = ${*$s}{fpos};
    if (!defined $whence || !$whence) {
        $fpos = $pos;
    } elsif ($whence == 1) {
        $fpos += $pos;
    } elsif ($whence == 2) {
        $fpos = ${*$s}{length} + $pos;
    } else {
        return undef;
    }
    return undef if $fpos < 0 || $fpos > ${*$s}{length};
    ${*$s}{fpos} = ($fpos ? $fpos : '0 but true');
    my $foff = ${*$s}{offset} + $fpos;
    return undef if CORE::sysseek($s, $foff, 0) != $foff;
    return ${*$s}{fpos};
}

package SoMat::File::Multi;

@SoMat::File::Multi::ISA = qw(SoMat::File);

sub open {
    my $s = shift;
    my ($basename, $mode) = @_;
#print STDERR "SoMat::File::Multi open(), basename=$basename\n";
    if ($basename !~ /.sic$/i) {
        bless $s, 'SoMat::File';
        return $s->open(@_);
    }

    ${*$s}{filename} = $basename;
    local $_;
    my @parts = map { s/\%/\%\%/g; $_ } $basename =~ /^(.*\D)(\d+)(\D+)$/;
    ${*$s}{mode} = $mode;
    ${*$s}{nameformat}
        = $parts[0] . '%.0' . length($parts[1]) . 'd' . $parts[2];
    ${*$s}{first} = $parts[1];
    ${*$s}{offsets} = [];
    ${*$s}{fh_cache} = {};
    ${*$s}{fh} = undef;
    ${*$s}{fh_idx} = -1;
    ${*$s}{fpos} = 0;

    my $off = 0;
    my $i = 0;
    while (1) {
        my $fname = $s->_filename($i);
#print STDERR "SoMat::File::Multi open() loop, fname=$fname, i=$i\n";
        last unless -f $fname;
        ${*$s}{offsets}->[$i] = $off;
        $off += -s $fname;
        ++$i;
    }
    return undef unless $i > 0;
    ${*$s}{lastfile} = $i - 1;
    ${*$s}{length} = ${*$s}{offsets}->[$i] = $off;
    return (${*$s}{lastfile} >= 0);
}

sub _seek {
    my ($s) = @_;

    my $fpos = ${*$s}{fpos};
    my $off = ${*$s}{offsets};
    my $i = ${*$s}{fh_idx};
    unless (defined ${*$s}{fh} && $i >= 0
        && $fpos >= $off->[$i] && $fpos < $off->[$i + 1]) {
        ${*$s}{fh} = undef;
        for $i (0 .. ${*$s}{lastfile}) {
            next unless $fpos >= $off->[$i] && $fpos < $off->[$i + 1];
            unless (exists ${*$s}{fh_cache}->{$i}) {
                # Limit the number of cached file handles to 4
                ${*$s}{fh_cache} = {} if scalar(keys %{${*$s}{fh_cache}}) > 4;
                my $fname = $s->_filename($i);
                my $fh = Symbol::gensym();
                $fh = undef unless CORE::open($fh, $fname);
                ${*$s}{fh_cache}->{$i} = $fh;
            }
            ${*$s}{fh} = ${*$s}{fh_cache}->{$i};
            ${*$s}{fh_idx} = $i;
            last;
        }
        $i = ${*$s}{fh_idx};
    }
    return undef unless defined ${*$s}{fh};
    my $fileoff = $fpos - $off->[$i];
    return undef unless CORE::seek(${*$s}{fh}, $fileoff, 0);
    my $roff = tell(${*$s}{fh});
    return undef if !defined $roff || $roff != $fileoff;
    return $fpos;
}

sub _filename {
    my ($s, $i) = @_;
    return sprintf ${*$s}{nameformat}, $i + ${*$s}{first};
}

sub close {
}

sub read {
    my ($s, undef, $len, $off) = @_;
    $len = length($_[1]) unless defined $len;
    $len = ${*$s}{length} - ${*$s}{fpos}
        if ${*$s}{fpos} + $len > ${*$s}{length};
    $off = 0 unless defined $off;
    my $toget = $len;
    my $fpos = ${*$s}{fpos};
    my $dpos = 0;
    substr($_[1], $off) = '' if $toget == 0; # special case
    while ($toget > 0) {
        my $seeked = $s->_seek;
        #die("seek failed: $seeked should have been $fpos") if $fpos != $seeked;
        return undef if $fpos != $seeked; # Seek failed
        my $got = CORE::read(${*$s}{fh}, $_[1], $toget, $off + $dpos);
        last unless defined $got;
        $toget -= $got;
        $dpos += $got;
        $fpos = (${*$s}{fpos} += $got);
    }
    #print "READ $dpos bytes from ${*$s}{fpos}\n";
    #my $data = $_[1];
    #$data =~ s/(.)/uc sprintf(" %02x", ord($1))/seg;
    #print "[$data ]\n";
    return $dpos || undef;
}

sub write {
    my ($s, $data, $len, $off) = @_;
    die "SoMat::File::Multi::write() not supported";
}

sub seek { # KLUDGE this does not do the right thing on error
    my ($s, $pos, $whence) = @_;
    my $fpos = ${*$s}{fpos};
    if (!defined $whence || !$whence) {
        $fpos = $pos;
    } elsif ($whence == 1) {
        $fpos += $pos;
    } elsif ($whence == 2) {
        $fpos = ${*$s}{length} + $pos;
    } else {
        return undef;
    }
    return undef if $fpos < 0 || $fpos > ${*$s}{length};
    return ${*$s}{fpos} = ($fpos ? $fpos : '0 but true');
}

1;
