
package SoMat::Network::Dispatch;
use strict;

sub new {
    my $class = shift;
    my $s = bless {}, ref($class) || $class;
    $s->{v_rd} = '';
    $s->{v_wr} = '';
    $s->{v_ex} = '';
    $s->{rd} = [];
    $s->{wr} = [];
    $s->{ex} = [];
    return $s;
}

sub _get_fd {
    my ($s, $fh) = @_;
    return fileno($fh) if ref($fh) && defined fileno($fh);
    return $fh;
}

sub add {
    my ($s, $mode, $fh, $sub, $data) = @_;
    my $fd = $s->_get_fd($fh);
    if ($mode eq 'W') {
        vec($s->{v_wr}, $fd, 1) = 1;
        push(@{$s->{wr}}, [$fd, $fh, $sub, $data]);
    } elsif ($mode eq 'R') {
        vec($s->{v_rd}, $fd, 1) = 1;
        push(@{$s->{rd}}, [$fd, $fh, $sub, $data]);
    }
    vec($s->{v_ex}, $fd, 1) = 1;
    push(@{$s->{ex}}, [$fd, $fh, $sub, $data]);
}

sub _remove_fd {
    my ($s, $ar, $fd) = @_;
    for my $i (0 .. @$ar - 1) {
        if ($ar->[$i]->[0] == $fd) {
            splice(@$ar, $i, 1);
            return;
        }
    }
}

sub remove {
    my ($s, $fh) = @_;
    my $fd = $s->_get_fd($fh);
    vec($s->{v_rd}, $fd, 1) = 0;
    vec($s->{v_wr}, $fd, 1) = 0;
    vec($s->{v_ex}, $fd, 1) = 0;
    $s->_remove_fd($s->{rd}, $fd);
    $s->_remove_fd($s->{wr}, $fd);
    $s->_remove_fd($s->{ex}, $fd);
}

sub dispatch {
    my ($s, $timeout) = @_;
    my $rd = $s->{v_rd};
    my $wr = $s->{v_wr};
    my $ex = $s->{v_wr};
    my $n = select($rd, $wr, $ex, $timeout);
    return unless $n;
    my $f;
    for $f (@{$s->{ex}}) {
        my ($fd, $fh, $sub, $data) = @$f;
        if (vec($ex, $fd, 1)) {
            die("Exception on FD $fd");
        }
    }
    for $f (@{$s->{rd}}) {
        my ($fd, $fh, $sub, $data) = @$f;
        if (vec($rd, $fd, 1)) {
            &$sub($fh, $data);
        }
    }
    for $f (@{$s->{wr}}) {
        my ($fd, $fh, $sub, $data) = @$f;
        if (vec($wr, $fd, 1)) {
            &$sub($fh, $data);
        }
    }
}

1;
