package SoMat::SIE::Perl;

use strict;
use warnings 'all';

use SoMat::File;
#use String::CRC32 ();
require String::CRC32;

#use Data::Dumper; $Data::Dumper::Indent = 1; $Data::Dumper::Sortkeys = 1;
#print "n = ", Data::Dumper::Dumper($n);

use Carp qw(cluck confess);
$SIG{__WARN__} = sub { confess @_; }; # cluck to just warn
$SIG{__DIE__} = sub { confess @_; };

my $SieBlockOverhead = 20; # Size, Sync, Group, <payload> , Checksum, Size
my $SiePreOverhead = 12; # Size, Sync, Group
my $SiePostOverhead = $SieBlockOverhead - $SiePreOverhead; # Checksum, Size
my $MagicValue = 0x51EDA7A0; # Squint to see "SIEDATA0"
my $MagicString = pack('N', $MagicValue);
my $Scale32 = 2.0 ** 32;

sub new {
    my $class = shift;
    my $s = bless {}, ref($class) || $class;
    $s->{decoder_debug} = 0;
    $s->{block_index} = [];
    $s->{tests} = {};
    push @{$s->{tags_stack}}, ($s->{stream_tags} = {});
    $s->_init_unpack;
    return $s;
}

sub IsThisFileThisType {
    my ($s, $fname) = @_;
    my $fh = SoMat::File->new($fname) or return undef;
    return defined $s->read_block($fh);
}

sub GetChannels {
    my ($s) = @_;
    my $ch = [ sort { $a <=> $b } keys %{$s->{channel_to_group}} ];
    return $ch;
}

sub Get {
    my ($s, $chan, $keyword) = @_;
    my $keywords = $s->{channels}->{$chan}->{keywords};
    return $keywords unless defined $keyword;
    return $keywords->{$keyword};
}

sub GetData {
    my ($s, $chan, $group) = @_;
    # KLUDGE nasty
    my @groups = (defined $group) ? ($group)
        : sort { $a <=> $b } keys %{$s->{channel_to_group}->{$chan}};

    delete $s->{channels}->{$chan}->{data}; # if defined $group; # HIDEOUS KLUDGE!

    for my $g (@groups) {
        for my $b (@{$s->{group_map}->{$g}}) {
            $s->{fh}->seek($b, 0) or die "Can't seek to $b";
            my ($group, $payload) = $s->read_block;
            $g == $group or die "Block $b is group $group, should be $g";
            $s->decode_payload($chan, $group, $payload);
        }
    }
#print "GetData = ", Dumper($s->{channels}->{$chan}->{data});
    return $s->{channels}->{$chan}->{data};
}

# Internal methods begin here

# Block reading

sub checksum {
    my ($s, $pre, $payload) = @_;
    my $check = String::CRC32::crc32($pre);
    return String::CRC32::crc32($payload, $check);
}

sub read_block {
    my ($s, $fh) = @_;
    $fh = $s->{fh} unless defined $fh;
    my $error;
#print "read_block A, pos = ", $fh->seek(0, 1), "\n";
    my ($pre, $payload, $post);
    $fh->read($pre, $SiePreOverhead) == $SiePreOverhead or return undef;
    my ($size, $group, $magic) = unpack('N3', $pre);
    unless ($magic == $MagicValue) {
        $error = "Bad magic";
        return undef 
    }
    my $plen = $size - $SieBlockOverhead;
    $fh->read($payload, $plen) == $plen or return undef;
    $fh->read($post, $SiePostOverhead) == $SiePostOverhead or return undef;
    my ($check, $size2) = unpack('N2', $post);
    if ($check != 0) {
        my $check2 = $s->checksum($pre, $payload);
        if ($check2 != $check) {
            print STDERR "SIE: read_block checksum mismatch\n";
        }
    }
    if ($size != $size2) {
        print STDERR "SIE: read_block size mismatch\n";
        return undef;
    }
    if ($error) {
        printf STDERR "SIE: read_block error %s:", $s->{error};
        printf STDERR " size=%s", (defined $size) ? $size : 'UNDEF';
        printf STDERR " group=%s", (defined $group) ? $group : 'UNDEF';
        printf STDERR " magic=%s",
            (defined $magic) ? sprintf("0x%X", $magic) : 'UNDEF';
        printf STDERR " checksum=%s",
            (defined $check) ? sprintf("0x%X", $check) : 'UNDEF';
        printf STDERR " size2=%s", (defined $size2) ? $size2 : 'UNDEF';
        printf STDERR "\n";
        return undef;
    }
    return ($group, $payload);
}

sub read_block_backward {
    my ($s) = @_;
    my $fh = $s->{fh};
#print "read_block_backward A, pos = ", $fh->seek(0, 1), "\n";
    defined $fh->seek(-4, 1) or return undef;
    my $buf;
    $fh->read($buf, 4) == 4 or return undef;
    my $size2 = unpack('N', $buf);
    defined $fh->seek(-$size2, 1) or return undef;
    my $pre;
    $fh->read($pre, $SiePreOverhead) == $SiePreOverhead or return undef;
    $fh->seek(-$SiePreOverhead, 1);
    my ($size, $group, $magic) = unpack('N3', $pre);
    return undef unless $size == $size2 && $magic == $MagicValue;
    return $group;
}

# Block finding and mapping stuff

sub search_blocks_backwards {
    my ($s) = @_;
    my $fh = $s->{fh};
    my $sblock = 8192;
    my $sbuf;
    my $pos = $fh->seek(0, 2); 
    return undef unless defined $pos;
    my @matches;
    while (1) {
        my $ssize = ($pos > $sblock) ? $sblock : $pos;
        $pos -= $ssize;
        $fh->seek($pos, 0);
        $fh->read($sbuf, $ssize) == $ssize or return undef;
        while ($sbuf =~ /$MagicString/sog) {
            push @matches, $pos + pos($sbuf) - $SiePreOverhead;
        }
        if (@matches) {
            return pop @matches; # KLUDGE maybe return all?
        }
        last if $pos <= 0;
        $pos += 3; # Overlap
    }
    return pop @matches;
}

sub map_file {
    my ($s, $fname) = @_;
    my $fh = $s->{fh};
    if (defined $fname) {
        $s->{filename} = $fname;
        $fh = $s->{fh} = SoMat::File->new($s->{filename})
            or die "Can't open $fname: $!";
    }
    my %map;
    $fh->seek(0, 2);
    while (1) {
#print "map_file LOOP, fpos = ", $fh->seek(0, 1), "\n";
        my $group = $s->read_block_backward;
        unless (defined $group) {
            last if $fh->seek(0, 1) == 0; # KLUDGE!!!
            warn "CORRUPTED FILE, doing search_blocks_backwards\n"; # KLUDGE
            my $pos = $s->search_blocks_backwards;
            $fh->seek($pos, 2);
            $group = $s->read_block_backward;
        }
        last unless defined $group;
        my $offset = $fh->seek(0, 1);
        $map{$offset} = $group;
        next unless $group == 1;

        # Decode Index Block (group 1)
        my $payload;
        ($group, $payload) = $s->read_block;
        my $lowest = $offset;
        for (my $pos = 0; $pos < length($payload); $pos += 12) {
             my ($offhi, $offlo, $gr) = unpack 'N3', substr($payload, $pos);
             $offset = $offhi * $Scale32 + $offlo;
             $map{$offset} = $gr;
             $lowest = $offset if $lowest > $offset;
        }
        $fh->seek($lowest, 0);
        last if $lowest == 0;
    }
    return \%map;
}

sub read_group {
    my ($s, $group) = @_;
    my $r = '';
    for my $off (@{$s->{group_map}->{$group}}) {
        $s->{fh}->seek($off, 0);
        my ($g, $payload) = $s->read_block;
        next if $g != $group; # KLUDGE CORRUPT BADNESS
        $r .= $payload;
    }
    return $r;
}

sub decode_payload { # This assumes ALL XML has already been processed
    my ($s, $ch, $group, $payload) = @_;

#print "decode_payload, s= ", Dumper($s);
    # A decoded group has v[0] .. v[m]. A channel has dimension d[0] .. d[n].
    # A v[k] can map to more than one dimension in a channel.
    # So we map:
    # {group_to_channel}->{$group}->{$channel}->{$dim} = $v
    my $chmap = $s->{group_to_channel}->{$group}->{$ch};
#print "decode_payload, chmap = ", Dumper($chmap);

    my %r; # Keyed by decoder;

#print "decode_payload, ch=$ch, group=$group\n";
    for my $dim (@{$s->{channels}->{$ch}->{dims}}) {
#print "  dim->{index}=$dim->{index}, dim->{v}=$dim->{v}\n";
        next unless exists $chmap->{$dim->{index}}
            && $chmap->{$dim->{index}} == $dim->{v};

        my $decoder = $dim->{decoder};
#print "  decoder = $decoder\n";
        die "No decoder for channel" unless defined $decoder;
        unless (defined $r{$decoder}) { # Skip if we've run it already
            $s->{p_buf} = $payload;
            $s->{p_pos} = 0;
            $s->{p_len} = length($payload);
            $r{$decoder} = eval { &{$s->{decoders}->{$decoder}}($s); };
        }
#print "r{$decoder} = ", Data::Dumper::Dumper($r{$decoder});
        my $r = $r{$decoder}->[$dim->{v}]; # KLUDGE check exists?
        die "decode_payload: Undefined result, ch=$ch, group=$group, dim=$dim->{index}, decoder=$decoder, pos=$s->{p_pos}, len=$s->{p_len}, source=$s->{decoder_source}->{$decoder}" unless defined $r;
        my $dest = $s->{channels}->{$ch}->{data}->[$dim->{index}] ||= [];

        if (defined $dim->{scale} || defined $dim->{offset}) {
            my $scale = $dim->{scale} || 1.0;
            my $offset = $dim->{offset} || 0.0;
            push @$dest, map { $_ * $scale + $offset } @$r;
        } elsif (defined $dim->{index_ch}) {
            my $map = $s->dimension_map($dim, $r);
            push @$dest, @$map;
        } else {
            push @$dest, @$r;
        }
    }
}

sub dimension_map {
    my ($s, $dim, $r) = @_;
#printf "dimension_map, index_ch=%s, group=%s, index_dim=%s\n",
#(defined $dim->{index_ch} ? $dim->{index_ch} : 'UNDEF'),
#(defined $dim->{group} ? $dim->{group} : 'UNDEF'),
#(defined $dim->{index_dim} ? $dim->{index_dim} : 'UNDEF');

    my $d = $s->GetData($dim->{index_ch}, $dim->{group});
#print Data::Dumper::Dumper($d);
    my %map;
    my $odim = $d->[$dim->{index_dim}];
    for my $i (@{$d->[0]}) {
        $map{$i} = $odim->[$i];
    }
    my @r = map { $map{$_} } @$r;
    return \@r;
}

sub read_file { # KLUDGE wrong name
    my ($s, $fname) = @_;
    $s->{fh} = SoMat::File->new($fname) or die "Can't open $fname: $!";
    my $idxmap = $s->map_file;
    $s->{group_map} = {};
    for my $off (sort { $a <=> $b } keys %$idxmap) {
        my $group = $idxmap->{$off};
        push @{$s->{group_map}->{$group}}, $off;
    }
    my $xml = $s->read_xml;
    if ($s->{XMLFile} && (my $fh = SoMat::File->new(">$s->{XMLFile}"))) {
        print $fh $xml;
    }
    $s->parse_xml($xml);
    for my $ch (keys %{$s->{xml_channels}}) {
        $s->assemble_channel_info($ch); # KLUDGE?
    }
}

# XML processing / interpretation

sub read_xml {
    my ($s) = @_;
    my $xml = $s->read_group(0);
    $xml .= "\n</sie>\n" unless $xml =~ /\<\/sie\>\s*$/s;
    return $xml;
}

sub assemble_channel_info {
    my ($s, $id) = @_;
    my $n = $s->merged_channel($id);

    # We have to walk the merged tree to get the tags
    my $tags = $s->{channels}->{$id}->{tags};
    push @{$s->{tags_stack}}, $tags;
    $s->walk_xml($n);
    pop @{$s->{tags_stack}};

    my $group = $n->attribute('group');
    $s->{channel_to_group}->{$id}->{$group} = 1 if defined $group;

    # InField-style keywords
    my $kw = $s->{channels}->{$id}->{keywords};

    # Merge test tags
    for my $tid (keys %{$s->{channels}->{$id}->{test}}) {
        while (my ($k, $v) = each %{$s->{tests}->{$tid}->{tags}}) {
            $kw->{$k} = $v;
        }
    }

    # Merge channel tags
    while (my ($k, $v) = each %$tags) {
        $kw->{$k} = $v;
    }
    if (defined (my $name = $n->attribute('name'))) {
        $kw->{ChanName} = $name;
    }

    # Assemble {dims} information, and do keyword stuff as well
    my $i = 0;
    while (my $dim = $n->getElement('dim', $i)) {
        my $idx = $dim->attribute('index');
        my $kwidx = $idx + 1;
        unless (defined $kw->{NumDims} && $kw->{NumDims} > $kwidx) {
            $kw->{NumDims} = $kwidx;
        }
        my $kwpfx = "Dim_$kwidx.";
        my $kwmap = {scale=>'Scale', offset=>'Offset'};
        my $gr = $group;
        my $cgr = $dim->attribute('group');
        $gr = $cgr if defined $cgr;

        my $chdim = ($s->{channels}->{$id}->{dims}->[$idx] ||= {});
        $chdim->{index} = $idx;
        my %f = (
            xform=>[qw(scale offset index_ch index_dim)],
            data=>[qw(decoder v)],
        );
        while (my ($tag, $attr) = each %f) {
            my $x = $dim->getElement($tag) or next;
            for my $a (@$attr) {
                my $av = $x->attribute($a);
                next unless defined $av;
                $chdim->{$a} = $av;
                $kw->{"$kwpfx$kwmap->{$a}"} = $av if $kwmap->{$a};
            }
        }
        for my $dt (@{$dim->getAllChildren}) {
            if ($dt->name eq 'tag') { # KLUDGE WRONG USE WALK
                $kw->{$kwpfx . $dt->attribute('id')} = $dt->text;
            }
        }
        if (defined $gr && defined $chdim->{v}) {
            $s->{group_to_channel}->{$gr}->{$id}->{$idx} = $chdim->{v};
            $s->{channel_to_group}->{$id}->{$gr} = 1;
        }
        ++$i;
    }

    $kw->{ChanNum} = $id + 1; # KLUDGE
    $kw->{BaseTime} = 0.0; # KLUDGE
    $kw->{SampleInc} = $kw->{'Dim_1.Scale'}
        if defined $kw->{'Dim_1.Scale'}; # KLUDGE
    $kw->{SampleRate} = 1.0 / $kw->{SampleInc}
        if defined $kw->{SampleInc}; # KLUDGE
}

sub merged_channel {
    my ($s, $id) = @_;
    my $na = $s->{xml_channels}->{$id};
    my $n = undef;
    for my $tn (@$na) {
        if (defined (my $base = $tn->attribute('base'))) {
            my $bc = $s->merged_channel($base);
            $tn = $bc->sie_merge($tn);
        }
        $n = ($n) ? $n->sie_merge($tn) : $tn->clone;
    }
    delete $n->{attr}->{base} if $n;
    return $n;
}

sub parse_xml {
    my ($s, $string) = @_;

    my $root = SoMat::SIE::Perl::XML->parse($string);
    $s->{xml_stack} = [];
    my $sie = $root->getElement('sie');
    $s->walk_xml($sie);
}

sub walk_xml {
    my ($s, $n) = @_;
    push @{$s->{xml_stack}}, $n;
    my $kids = $n->getAllChildren;
    for my $k (@$kids) {
        my $reader = '_xml_' . $k->name;
        die("Unrecognized XML tag " . $k->name) unless $s->can($reader);
        $s->$reader($k);
    }
    pop @{$s->{xml_stack}};
}

sub _xml_tag {
    my ($s, $n) = @_;
    my $id = $n->attribute('id');
    my $v;
    if (defined (my $group = $n->attribute('group'))) {
        my $format = $n->attribute('format');
        # KLUDGE does nothing with "decoder" attribute
        $v = $s->read_group($group); # KLUDGE bad in general - too huge
    } else {
        $v = $n->text;
    }
#print "_xml_tag $s->{tags_stack}->[-1] $id => $v\n";
    $s->{tags_stack}->[-1]->{$id} = $v;
}

sub _xml_ch {
    my ($s, $n) = @_;
    my $id = $n->attribute('id');
    push @{$s->{xml_channels}->{$id}}, $n;
    unless ($s->{channels}->{$id}) {
        $s->{channels}->{$id} = {
            tags=>{}, dims=>[], data=>[], test=>{}, keywords=>{}
        };
    }
    my $parent = $s->{xml_stack}->[-1];
    if ($parent->name eq 'test') {
        $s->{channels}->{$id}->{test}->{$parent->attribute('id')} = 1;
    }
    push @{$s->{tags_stack}}, $s->{channels}->{$id}->{tags};
    $s->walk_xml($n);
    pop @{$s->{tags_stack}};
}

sub _xml_dim {
    # Do nothing (yet?)
}

sub _xml_test {
    my ($s, $n) = @_;
    my $id = $n->attribute('id');
    push @{$s->{tags_stack}}, ($s->{tests}->{$id}->{tags} ||= {});
    $s->walk_xml($n);
    pop @{$s->{tags_stack}};
}

sub _xml_decoder {
    my ($s, $n) = @_;
    my $id = $n->attribute('id');
    $s->{_d_vars} = {};
    $s->{_d_max_v} = -1;
#print STDERR $n->toString(1);
    my $code = $s->_compile_decoder($n); # KLUDGE check result
    my $o = "sub {\n";
    $o .= "my \@r;\n";
    for my $dec (sort keys %{$s->{_d_vars}}) {
        $o .= "my \$$dec = 0;\n";
    }
    $o .= "\n" . $code;
    $o .= "return \\\@r;\n";
    $o .= "}\n";
#print STDERR $n->toString(1);
#print STDERR "DECODER ($id):\n$o\n";
    my $sub = eval $o;
    unless (defined $sub) {
        warn "Error defining decoder: $@";
    }
    $s->{decoders}->{$id} = $sub;
    $s->{decoder_source}->{$id} = $o;
}

# Decoder compiler

sub _compile_decoder {
    my ($s, $dec, $o) = @_;
    $o = '' unless defined $o;
    my $ops = $dec->getAllChildren;
    for my $op (@$ops) {
        my $name = $op->name;
        next if $name eq '!--'; # Comment
        my $sub = "_comp_$name";
        die("unrecognized tag: $name") unless $s->can($sub);
        $o = $s->$sub($op, $o);
    }
    return $o;
}

# Decoder compiler operator handlers

sub _comp_if {
    my ($s, $op, $o) = @_;
    my $cond = $s->_deexpr($op->attribute('condition'));
    die "if: no condition defined" unless defined $cond;
    $o .= "if ($cond) {\n";
    $o = $s->_compile_decoder($op, $o);
    $o .= "}\n";
    return $o;
}

sub _comp_loop {
    my ($s, $op, $o) = @_;
    my $begin = '';
    my $next = '';
    my $cond = '$_[0]->{p_pos} < $_[0]->{p_len}';
    my $inc = $s->_deexpr($op->attribute('increment') || 1);
    my $start = $s->_deexpr($op->attribute('start'));
    my $end = $s->_deexpr($op->attribute('end'));
    if (my $var = $s->_get_var($op)) {
        $next = "\$$var += $inc";
        $begin = "\$$var = $start" if defined $start;
        $cond = "\$$var < $end" if defined $end;
    } elsif (defined $start) {
        die "loop: start = $start with no var";
    } elsif (defined $end) {
        die "loop: end = $end with no var";
    }
    $o .= "for ($begin; $cond; $next) {\n";
    $o = $s->_compile_decoder($op, $o);
    $o .= "}\n";
    return $o;
}

sub _comp_read {
    my ($s, $op, $o) = @_;
    my $var = $s->_get_var($op);
    my $type = $op->attribute('type') or warn "No type for $var";
    my $size = $s->_deexpr($s->_get_size($op)) or warn "No size for $var";
    my $endian = $op->attribute('endian');
    my $unpack;
    if ($type eq 'raw') {
        $unpack = 'a*';
    } elsif ($size =~ /^\d+$/) {
        $unpack = $s->_get_unpack($endian, $type, $size * 8);
    }
    die($op->toString . ": No unpack for $var ($endian, $type $size)")
        unless defined $unpack;
    my $uop = ($unpack =~ /_/) ? '$_[0]->_unpack(' : 'unpack(';
    $o .= "\$$var = $uop'$unpack', \$_[0]->_read($size));\n";
    $o .= "print qq(read: $var=\$$var\\n);\n" if $s->{decoder_debug};
    return $o;
}
        
sub _comp_sample {
    my ($s, $op, $o) = @_;
#print "_sample: max_v=$s->{_d_max_v}\n";
    die "Error: _d_max_v = $s->{_d_max_v}" if $s->{_d_max_v} < 0;
    $o .= "print qq(sample: [);\n" if $s->{decoder_debug};
    for my $i (0 .. $s->{_d_max_v}) {
        $o .= "push \@{\$r[$i]}, \$v$i;\n";
    $o .= "print qq(, );\n" if $i && $s->{decoder_debug};
    $o .= "print qq(\$v$i);\n" if $s->{decoder_debug};
    }
    $o .= "print qq(]\\n);\n" if $s->{decoder_debug};
    return $o;
}

sub _comp_seek {
    my ($s, $op, $o) = @_;
    my $from_map = {start=>0, current=>1, end=>2};
    my $from = $op->attribute('from');
    my $offset = $s->_deexpr($op->attribute('offset'));
    die "Error: invalid offset" unless defined $offset;
    die "Error: invalid seek from=$from" unless defined $from_map->{$from};
#print "_seek: from=$from, offset=$offset\n";
    $o .= "\$_[0]->_seek($from_map->{$from}, ($offset));\n";
    return $o;
}

sub _comp_set {
    my ($s, $op, $o) = @_;
    my $var = $s->_get_var($op);
    my $val = $op->attribute('value');
#print "_set: var=$var, val=$val\n";
    $val = $s->_deexpr($val);
    $o .= "\$$var = $val;\n";
    return $o;
}

# Decoder compiler support code

sub _read {
    my ($s, $bytes) = @_;
#print "_read: bytes=$bytes, pos=$s->{p_pos}, len=$s->{p_len}\n";
    die if $s->{p_pos} < 0
        || $s->{p_pos} + $bytes > $s->{p_len}; # Busts out of eval
    my $got = substr($s->{p_buf}, $s->{p_pos}, $bytes);
    $s->{p_pos} += $bytes;
#print "_read: ok, pos=$s->{p_pos}\n";
    return $got;
}

sub _seek {
    my ($s, $from, $offset) = @_;
    # <seek> only repositions the pointer. Nothing bad should happen
    # from an out of bounds <seek> until the next <read> occurs.
    if ($from == 0) {
        $s->{p_pos} = $offset;
    } elsif ($from == 1) {
        $s->{p_pos} += $offset;
    } elsif ($from == 2) {
        $s->{p_pos} = $s->{p_len} + $offset;
    }
}

sub _init_unpack {
    my ($s) = @_;
    $s->{Endian} = (pack('L', 1) eq "\x01\x00\x00\x00") ? 'L' : 'B';
    $s->{EndianMap} = {
        big=>'B',
        little=>'L',
    };
    my $m = { # Initialize with portable stuff
        B_uint_32=>'N',
        B_uint_16=>'n',
        B_uint_8=>'C',
        B_int_8=>'c',

        L_uint_32=>'V',
        L_uint_16=>'v',
        L_uint_8=>'C',
        L_int_8=>'c',
    };
    # Floating point is all IEEE 754
    if ($s->{Endian} eq 'L') {
        # Little-endian
        $m->{L_int_32} = 'l';
        $m->{L_int_16} = 's';
        $m->{L_float_32} ='f';
        $m->{L_float_64} ='d';
    } else {
        # Big-endian
        $m->{B_int_32} = 'l';
        $m->{B_int_16} = 's';
        $m->{B_float_32} ='f';
        $m->{B_float_64} ='d';
    }
    $s->{UnpackMap} = $m;
    # Anything not in portable section of UnpackMap should be here
    $m = {};
    $m->{L_uint_64} = sub {
        my ($lo, $hi) = unpack('VV', $_[0]);
        return $lo + $Scale32 * $hi;
    };
    $m->{B_uint_64} = sub {
        my ($hi, $lo) = unpack('NN', $_[0]);
        return $lo + $Scale32 * $hi;
    },
    $m->{L_int_64} = sub {
        my ($lo, $hi) = unpack('VV', $_[0]);
        if ($hi & 0x80000000) {
            $hi ^= 0xFFFFFFFF;
            $lo ^= 0xFFFFFFFF;
            return -1.0 - $lo - $Scale32 * $hi;
        }
        return $lo + $Scale32 * $hi;
    };
    $m->{B_int_64} = sub {
        my ($hi, $lo) = unpack('NN', $_[0]);
        if ($hi & 0x80000000) {
            $hi ^= 0xFFFFFFFF;
            $lo ^= 0xFFFFFFFF;
            return -1.0 - $lo - $Scale32 * $hi;
        }
        return $lo + $Scale32 * $hi;
    };
    # For the rest, we're opposite-endian, otherwise we couldn't get here
    $m->{B_int_32} = $m->{L_int_32} = sub {
        return unpack('l', reverse(substr($_[0], 0, 4)))
    };
    $m->{B_int_16} = $m->{L_int_16} = sub {
        return unpack('s', reverse(substr($_[0], 0, 2)))
    };
    $m->{B_float_32} = $m->{L_float_32} = sub {
        return unpack('f', reverse(substr($_[0], 0, 4)))
    };
    $m->{B_float_64} = $m->{L_float_64} = sub {
        return unpack('d', reverse(substr($_[0], 0, 8)))
    };
    $s->{UnpackSubs} = $m;
}

sub _unpack {
    my ($s, $type, $v) = @_;
    return &{$s->{UnpackSubs}->{$type}}($v);
}

sub _get_size {
    my ($s, $x) = @_;
    my $size = $x->attribute('octets');
    return $size if defined $size;
    $size = $x->attribute('bits');
    die "No size" unless defined $size;
    die "sub-byte sizes not (yet) supported" if $size % 8;
    return $size / 8;
}

sub _get_unpack {
    my ($s, $endian, $type, $size) = @_;
    my $e = $s->{EndianMap}->{$endian} or die "Unknown endian: $endian";
    my $t = "${e}_${type}_${size}";
    return $s->{UnpackMap}->{$t} || $t;
}

sub _deexpr {
    my ($s, $e) = @_;
    return undef unless defined $e;
    return $e unless $e =~ /^\{(.*?)\}$/;
    $e = $1;
    return $e;
}

sub _get_var {
    my ($s, $op) = @_;
    return undef unless my $var = $op->attribute('var');
    $s->{_d_vars}->{$var} = 1;
    if ($var =~ /^v(\d+)$/) {
        $s->{_d_max_v} = $1 if $s->{_d_max_v} < $1
    }
    return $var;
}

# File validation / recovery stuff

sub check_file {
    my ($s, $fname) = @_;
    $s->{fh} = SoMat::File->new($fname) or die "Can't open $fname: $!";
    my $filesize = -s $s->{fh};
    print "check_file[0], size=$filesize bytes\n";
    my %map;
    my %idxmap;
    while (1) {
        my $off = $s->{fh}->seek(0, 1);
        my ($group, $payload) = $s->read_block;
        unless (defined $group) {
            last if $off >=  $filesize;
            die "Bad block at offset $off\n";
        }
        my $len = length($payload);
        printf "check_file[1], o=$off, size=%d, group=$group, payload=%d bytes, next=%d\n", $len + $SieBlockOverhead, $len, $off + $len + $SieBlockOverhead;
        $map{$off} = length($payload) + $SieBlockOverhead;
        if ($group == 1) {
            my $i = 0;
            for (my $pos = 0; $pos < length($payload); $pos += 12) {
                my ($offhi, $offlo, $gr) = unpack 'N3', substr($payload, $pos);
                my $offset = $offhi * $Scale32 + $offlo;
                print "  map[$i]=> group=$gr, offset=$offset\n";
                $idxmap{$offset} = $gr;
                ++$i;
            }
        }
    }
    $map{$s->{fh}->seek(0, 2)} = 0;
    my @offs = sort { $a <=> $b } keys %map;
    while (@offs) {
        my $o = shift @offs;
        last unless @offs;
        my $next = $o + $map{$o};
        #print "check_file, o = $o, next = $next\n";
        if ($next != $offs[0]) {
            printf "Unaccounted space at $next: %d bytes!\n", $offs[0] - $next;
        }
    }
    #my $idxmap = $s->map_file;
    for my $off (sort { $a <=> $b } keys %idxmap) {
        unless (exists $map{$off}) {
            print "map_file wrongly claims block at $off\n";
            next;
        }
        delete $map{$off};
    }
    for my $off (sort { $a <=> $b } keys %map) {
        print "map_file missed block at $off\n" if $map{$off};
    }
}

# SIE File writing

sub create {
    my ($s, $fname) = @_;
    $s->{offset} = 0;
    return $s->{fh} = SoMat::File->new(">$fname");
}

sub close {
    my ($s) = @_;
    $s->write_index_block or return undef;
    return $s->{fh}->close;
}

sub write_block {
    my ($s, $group, $payload) = @_;
    my $size = length($payload) + $SieBlockOverhead;
    my $pre = pack('N3', $size, $group, $MagicValue);
    my $check = $s->checksum($pre, $payload); # KLUDGE conditional?
    my $post = pack('N2', $check, $size);
    my $block = $pre . $payload . $post;
    $s->{fh}->write($block) == $size or die "Error writing block: $!";
    my $off = $s->{offset};
    $s->{offset} += $size;
    push @{$s->{block_index}}, $off, $group;
    return ($off, $group);
}

sub write_index_block {
    my ($s) = @_;
    my $idx = $s->{block_index};
    return undef unless @$idx;
    my $pkt = '';
    while (@$idx) {
        my ($offset, $group) = splice(@$idx, 0, 2);
        my $offhi = int($offset / $Scale32);
        my $offlo = int($offset - $offhi * $Scale32);
        $pkt .= pack('N3', $offhi, $offlo, $group);
    }
    my @r = $s->write_block(1, $pkt) or return undef;
    $s->{block_index} = [];
    return @r;
}

package SoMat::SIE::Perl::XML;

# Note:
# * The API is very similar to XML::Mini but this is all original code
# * This has NOTHING extra that isn't needed for SIE XML.

sub root {
    my ($s) = @_;
    return $s->new('');
}

sub new {
    my ($class, $name) = @_;
    my $s = bless {}, ref($class) || $class;
    $s->{name} = $name;;
    $s->{attr} = {};;
    $s->{kids} = [];
    $s->{text} = undef;
    return $s;
}

sub name {
    my ($s, $val) = @_;
    return $s->{name} unless defined $val;
    $s->{name} = $val;
}

sub attribute {
    my ($s, $name, $val) = @_;
    return $s->{attr}->{$name} unless defined $val;
    $s->{attr}->{$name} = $val;
}

sub text {
    my ($s, $val) = @_;
    return $s->{text} unless defined $val;
    $s->{text} = $val;
}

sub getElement {
    my ($s, $name, $n) = @_;
    $n = 0 unless defined $n;
    for my $k (@{$s->{kids}}) {
        return $k if $k->{name} eq $name && !$n--;
    }
    return undef;
}

sub createChild {
    my ($s, $name) = @_;
    my $k = $s->new($name);
    push @{$s->{kids}}, $k;
    return $k;
}

sub getAllChildren {
    my ($s) = @_;
    return $s->{kids};
}

sub parse {
    my ($s, $buf) = @_;

    my $root = $s->root;
    my @stack = ($root);

    while (1) {
        if ($buf =~ /^\s*\<([^\>]+)\>\s*(.*)$/s) {
            my $e = $1;
            $buf = $2;
            if ($e =~ /^\?.*\?$/s) {
                next; # KLUDGE
            }
            if ($e =~ /^!--(.*)--$/s) {
                next;
            }
            if ($e =~ /^\/(\w+)$/s) {
                pop @stack;
                next;
            }
            if (my ($tag, $att, $end)
                = $e =~ /^(\w+)((?:\s+\w+\=\".*?\")*)\s*(\/)?$/s) {
                $end = '' unless defined $end;
#print STDERR "XML: tag=$tag, att=$att, end=$end\n";
                my $k = $stack[-1]->createChild($tag);
                push(@stack, $k) unless $end;
                while ($att =~ /\G\s*(\w+)\=\"(.*?)\"/sg) {
                    $k->attribute($1, $2);
                }
                next;
            }
            die "E: $e\n";
        }
        if ($buf =~ /^\s*([^\<]+)(.*)$/s) {
            $stack[-1]->text($1);
            $buf = $2;
            next;
        }
        last unless length $buf;
    }
    return $root;
}

sub toString {
    my ($s) = @_;
    my $name = defined $s->{name} ? $s->{name} : 'UNDEF';
    my $attr = '';
    for my $a (sort keys %{$s->{attr}}) {
        $attr .= qq( $a="$s->{attr}->{$a}");
    }
    return "<$name$attr>$s->{text}</$name>\n" if defined $s->{text};
    return "<$name$attr/>\n" unless @{$s->{kids}};
    my $contents = defined $s->{text} ? $s->{text} : '';
    for my $k (@{$s->{kids}}) {
        $contents .= $k->toString;
    }
    return "<$name$attr>\n$contents</$name>\n";
}

# XML merging stuff

sub clone {
    my ($s) = @_;
    my $ref = ref($s);
    return $s unless $ref;
    my ($type) = $s =~ /\=([A-Z]+)/;
    $type = $ref unless defined $type;
    my $r;
    if ($type eq 'HASH') {
        $r = {};
        while (my ($k, $v) = each %$s) {
            $r->{$k} = clone($v);
        }
    } elsif ($type eq 'ARRAY') {
        $r = [];
        for my $v (@$s) {
            push @$r, clone($v);
        }
    } else {
        die "Can't clone $ref";
    }
    bless($r, $ref) if $ref ne $type;
    return $r;
}

sub _overlay_kids {
    my ($m, $n, $r, $get_id, $submerge) = @_;
    my %r;
    for my $k (@{$n->{kids}}) {
        my $id = &$get_id($k);
        $r{$id} = $k if defined $id;
    }
    for my $k (@{$m->{kids}}) {
        my $id = &$get_id($k);
        next unless defined $id;
        if (defined $submerge) {
            $r{$id} = &$submerge($k, $r{$id});
        } else {
            $r{$id} = $k unless exists $r{$id};
        }
    }
    for my $id (sort keys %r) {
        push @{$r->{kids}}, clone($r{$id});
    }
}

sub sie_merge {
    my ($s, $n) = @_;
 
    return $s unless defined $n;
    return undef unless $s->{name} eq $n->{name};

    my $r = SoMat::SIE::Perl::XML->new;
    $r->{name} = $s->{name};
    while (my ($k, $v) = each %{$n->{attr}}) {
        $r->{attr}->{$k} = $v;
    }
    while (my ($k, $v) = each %{$s->{attr}}) {
        $r->{attr}->{$k} = $v unless defined $r->{attr}->{$k};
    }
    $r->{text} = (defined $n->{text}) ? $n->{text} : $s->{text};
    
    $s->_overlay_kids($n, $r, sub {
        $_[0]->name eq 'tag' ? $_[0]->attribute('id') : undef;
    });

    if ($s->name eq 'ch') {
        $s->_overlay_kids($n, $r, sub {
                $_[0]->name eq 'dim' ? $_[0]->attribute('index') : undef
            }, sub {
                $_[0]->sie_merge($_[1])
            }
        );
    } elsif ($s->name eq 'dim') {
        $s->_overlay_kids($n, $r, sub {
            $_[0]->name =~ /^(data|xform)$/ ? $_[0]->name : undef
        });
    }
    return $r;
}

1;
