package SoMat::eDAQ::Hardware::Eeprom;

use strict;
use Symbol;

my $EEPROM_ECN_OFFSET = 0x1e00;
my $MAX_SLOT_ECNS = 60;
my $ECN_SIZE = 16 + 8 + 4 + ((4 + 4) * $MAX_SLOT_ECNS);

sub read_eeprom_ecns {
    my ($fh) = @_;
    my ($buf, $buf2);

    sysseek($fh, $EEPROM_ECN_OFFSET, 0);
    sysread($fh, $buf, 16 + 8 + 4) || die "bad read";
    my ($foo, $foo2, $count) = unpack("Z16Z8V", $buf);
    if ($count <= 0 || $count > $MAX_SLOT_ECNS) {
        $count = 0;
        return $buf;
    }
    sysread($fh, $buf2, (4 + 4) * $count) || die "bad read";
    return $buf . $buf2;
}

sub interpret_ecns {
    my ($buf) = @_;
    my %ret;
    ($ret{header}, $ret{version}, $ret{count}) = unpack("Z16Z8V", $buf);
    if ($ret{count} > $MAX_SLOT_ECNS) {
        $ret{count} = 0;
        return \%ret;
    }
    for (my $i = 0; $i < $ret{count}; $i++) {
        my ($date, $year, $month, $day);
        ($ret{ecns}->[$i]->{number}, $date) =
            unpack("VV", substr($buf, 16 + 8 + 4 + 8 * $i));
        $year = $date % 100;
        $date = int($date / 100);
        $day = $date % 100;
        $month = int($date / 100);
        if ($year < 90) {
            $year += 2000;
        } else {
            $year += 1900;
        }
        $ret{ecns}->[$i]->{date} = $year * 10000 + $month * 100 + $day;
    }
    return \%ret;
}

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($device) = @_;
    my $self  = {};
    $self->{filehandle} = gensym;
    $self->{device} = $device;
    $self->{type} = "Eeprom";
    bless ($self, $class);
    if (!open($self->{filehandle}, "<$device")) {
        return undef;
    }
    return $self;
}

sub DESTROY {
    my $self = shift;
    close($self->{filehandle});
}

sub get_serial {
    my $self = shift;
    my $serial;
    sysseek($self->{filehandle}, 0, 0);
    sysread($self->{filehandle}, $serial, 13);
    return $serial;
}

sub get_ecns {
    my $self = shift;
    return interpret_ecns(read_eeprom_ecns($self->{filehandle}));
}

1;

