package SoMat::eDAQ::Hardware::Expmod;

use strict;
use Symbol;

my $MAX_PACKET_SIZE = 1024;
my $PKT_READ_TIMEOUT = 5.0; # 5 seconds

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($device) = @_;
    my $self  = {};
    $self->{type} = "Expmod";
    $self->{device} = $device;
    $self->{filehandle} = gensym;
    $self->{port} = 0xff00;  # default
    bless ($self, $class);
    if (!open($self->{filehandle}, "+<$device")) {
        return undef;
    }
    return $self;
}

sub DESTROY {
    my $self = shift;
    close($self->{filehandle});
}

sub set_port {
    my $self = shift;
    $self->{port} = shift;
}

sub set_timeout {
    my $self = shift;
    $PKT_READ_TIMEOUT = shift;
}

sub send_packet {
    my $self = shift;
    my ($packet, $port) = @_;
    $port ||= $self->{port};
#print STDERR "sending $packet\n";
    return syswrite($self->{filehandle}, 
        pack("SSa*", (length($packet) + 2, $port, $packet)));
}

sub recv_packet {
    my $self = shift;
    my $port = shift || $self->{port};
    my $timeout = $PKT_READ_TIMEOUT;
    my ($length, $packedlength, $packet);
    my ($rin, $rout);

    $rin = '';
    vec($rin, fileno($self->{filehandle}), 1) = 1;
    
    while (1) {
#printf STDERR "trying fd %d\n", fileno($self->{filehandle});
        my ($nfound, $timeleft) = select($rout = $rin, undef, undef, $timeout);
        if (!vec($rout, fileno($self->{filehandle}), 1)) {
#print STDERR "read bit not set, timeleft = $timeleft\n";
            return undef;
        }
        sysread($self->{filehandle}, $packedlength, 2); # read packet size
        $length = unpack("S", $packedlength);
        my $count = sysread($self->{filehandle}, $packet, $length + ($length & 1));
        if ($count != $length + ($length & 1)) {
#print STDERR "count $count != length $length\n";
            return undef;
        }
#printf STDERR "got response on port 0x%04x (want 0x%04x): %s (%s)\n", unpack("v", substr($packet, 0, 2)), $port, substr($packet, 2, $length - 2), unpack("H*", $packet);
        if (pack("v", $port) eq substr($packet, 0, 2)) {
#print STDERR "MATCH!\n";
            return substr($packet, 2, $length - 2);
        }
        $timeout = $timeleft;  # KLUDGE unportable?
        if ($timeout <= 0) {
            return undef;
        }
    }
}

sub recv_cmd_response {
    my $self = shift;
    my ($response, $packet);
    
    while ( ($packet = $self->recv_packet()) ) {
        $response .= $packet . "\n";
        last if (substr($packet, 0, 5) eq 'ERRNO');
    }

#print STDERR "recv_cmd_response got: $response";

    return $response;
}

sub get_named_variable {
    my $self = shift;
    my ($name, $other) = @_;
    $self->send_packet("GET $other NAME=$name");
    my $result = $self->recv_cmd_response();
#print STDERR "result: $result";
    if ($result =~ /$name=(\S+)/i) {
        return $1;
    }
    return undef;
}

sub get_serial {
    my $self = shift;
    my $sn = $self->get_named_variable("SERIAL_NUMBER");
    if ($self->get_named_variable("APPLICATION_IMAGE_CORRUPT") ne "FALSE" ||
        $self->get_named_variable("PRIMARY_FPGA_IMAGE_CORRUPT") ne "FALSE" ||
        $self->get_named_variable("PRIMARY_FPGA_IMAGE_LOAD_FAILURE") ne "FALSE") {
        $sn = "X$sn";
    }
    return $sn;
}

sub get_version {
    my $self = shift;
    my $version = $self->get_named_variable("FW_VERSION");
    $version =~ tr/A-Z/a-z/;
    return $version;
}

sub get_build {
    my $self = shift;
    my $version = $self->get_named_variable("FW_BUILD_NUMBER");
    $version =~ tr/A-Z/a-z/;
    return $version;
}

sub get_ecns {
    my $self = shift;
    $self->send_packet("ECN_READ");
    my $result = $self->recv_cmd_response();
    my %ret;
    $ret{count} = 0;
    foreach my $line (split("\n", $result)) {
        if ($line =~ /ECN_NUMBER=(\S+)\s+ECN_DATE=(\S+)/i) {
            $ret{ecns}->[$ret{count}]->{number} = $1;
            $ret{ecns}->[$ret{count}]->{date} = $2;
            $ret{count}++;
        }
    }
    return \%ret;
}

1;
