package SoMat::eDAQ::Hardware::Update;

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    open($s->{dev}, '+<', $s->{device_name})
        or die "Can't open device $s->{device_name}: $!";
    return $s;
}

sub hex {
    my ($s, $buf) = @_;
    my $hex = unpack('H*', $buf);
    $hex =~ s/(..)/$1 /g;
    return $hex;
}

sub debug {
    my ($s, $fmt, @args) = @_;
    return unless $s->{debug};
    my $msg = sprintf $fmt, @args;
    print $msg;
}

sub message {
    my ($s, %p) = @_;
    my $p = \%p;
    $p->{type} ||= 'message';
    while (my ($k, $v) = each %{$s->{progress}}) {
        $p->{$k} ||= $v;
    }
    $p->{time} = time - $p->{base_time};
    $s->emit_message($p);
}

sub init_progress {
    my ($s, $todo) = @_;
    $s->{progress} = {
        type => 'progress',
        name => $s->{name},
        serial => $s->{serial},
        done => 0,
        todo => $todo,
        progress => 0,
        steps => 100, # KLUDGE?
        next => 0, # KLUDGE?
        base_time => time(),
        message => '',
    };
}

sub progress_update {
    my ($s, $done) = @_;
    $s->{progress}{done} = $done;
}

sub progress {
    my ($s, $fmt, @args) = @_;
    my $msg = sprintf $fmt, @args;
    my $p = $s->{progress};
    my $out = ($p->{message} ne $msg);
    $p->{message} = $msg;
    my $time = time - $p->{base_time};
    $out = 1 if $p->{time} != $time; # Update at least at 1 Hz
    $p->{time} = $time;
    unless ($out) {
        if ($p->{done} >= $p->{next}) {
            $out = 1;
            $p->{next} = (1 + int(($p->{done} / $p->{todo}) * $p->{steps}))
                * ($p->{todo} / $p->{steps});
        }
    }
    return unless $out;
    $p->{progress} = 100.0 * $p->{done} / $p->{todo};
    $s->emit_message($p);
}

sub emit_message {
    my ($s, $p) = @_;
    my @out;
    while (my ($k, $v) = each %$p) {
        $v =~ s/([\\"\x00-\x19])/sprintf('\\x%02X', ord($1))/seg;
        push @out, qq{"$k"=>"$v"};
    }
    print "{", join(",", @out), "}\n";
}

1;
