
package SoMat::eDAQ::SRam;

use strict;
use vars qw($VERSION);
$VERSION = sprintf("%d.%02d", q$Revision: GIT $ =~ /(\d+)\.(\d+)/);

require SoMat::File;
require SoMat::eDAQ::Structurizer;

sub new {
    my $class = shift;
    my $s = bless {}, ref $class || $class;
    $s->{filename} = shift || '/dev/ramdisk';
    my $flags = shift || "r";
    my $mode = ($flags =~ /w/) ? SoMat::File::O_RDWR() : SoMat::File::O_RDONLY();
    $s->{fh} = SoMat::File->new($s->{filename}, $mode) or return undef;
    my $szr = $s->{szr} = SoMat::eDAQ::Structurizer->new($s->{fh});
    $szr->read('RD_ROOT_OBJ', 0);
    $szr->read('RD_NODE_OBJ');
    #my $off = $szr->get('next_ofs'); # KLUDGE Do this to skip "!" entry
    my $off = $szr->get('self_ofs');
    $s->{dir} = [];
    $s->{rd_node_size} = $szr->size;
    while ($off) {
        $szr->read('RD_NODE_OBJ', $off);
        my $date = sprintf "%.04d-%.02d-%.02d %.02d:%.02d:%.02d",
            $szr->get('TimeDate', 'year'),
            $szr->get('TimeDate', 'month'),
            $szr->get('TimeDate', 'day'),
            $szr->get('TimeDate', 'hour'),
            $szr->get('TimeDate', 'minute'),
            $szr->get('TimeDate', 'second');
        my $self_ofs = $szr->get('self_ofs');
        push(@{$s->{dir}}, { 'name' => $szr->get('name'), 'date' => $date,
            'rd_node_offset' => $self_ofs,
            'offset' => $self_ofs + $s->{rd_node_size},
            'size' => $szr->get('size')});
        $off = $szr->get('next_ofs');
#print STDERR "Got $s->{dir}->[-1]->{name}, size = $s->{dir}->[-1]->{size}, next = $off\n";
        last if $off == $self_ofs; # Avoid infinite loops from buggage
    }
    return $s;
}

sub DESTROY {
    my ($s) = @_;
    close($s->{fh}) if defined $s->{fh};
}

sub get_fh {
    my ($s) = @_;
    return $s->{fh};
}

sub get_dir {
    my ($s) = @_;
    return (wantarray) ? @{$s->{dir}} : $s->{dir};
}

sub rewind {
    $_[0]->{pos} = 0;
}

sub readdir {
    my ($s) = @_;
    return undef unless $s->{off};
    if (wantarray) {
        local $_;
        return map { $_->{name} } @{$s->{dir}};
    }
    return undef if $s->{pos} >= @{$s->{dir}};
    return $s->{dir}->[($s->{pos})++]->{name};
}

sub create {
    my ($s) = @_;
    my $last = $s->{dir}->[-1];
}

sub get_rd_node {
    my ($s, $name) = @_;
    for my $node (@{$s->{dir}}) {
        return $node if $node->{name} eq $name;
    }
    return undef;
}

sub rd_node_put {
    my ($s, $field, @val) = @_;
    my $szr = $s->{szr};
    my $map = {
        'self_ofs' => ['self_ofs', 'self2_ofs', 'self3_ofs'],
        'next_ofs' => ['next_ofs', 'next2_ofs', 'next3_ofs'],
        'name' => ['name', 'name2', 'name3'],
        'size' => ['size', 'size2', 'size3']
        # checksum, TimeDate not tripled
    };
    if (exists $map->{$field}) {
        for my $k (@{$map->{$field}}) {
            $szr->put($k, @val); # Kinda icky # KLUDGE errors?
        }
    } else {
        $szr->put($field, @val); # Kinda icky, again # KLUDGE errors?
    }
}

sub set_rd_node {
    my ($s, $node, $field, @val) = @_;
    $node = $s->get_rd_node($node) unless ref($node) eq 'HASH';
    my $szr = $s->{szr};
    $szr->read('RD_NODE_OBJ', $node->{rd_node_offset});
    my $map = {
        'self_ofs' => ['self_ofs', 'self2_ofs', 'self3_ofs'],
        'next_ofs' => ['next_ofs', 'next2_ofs', 'next3_ofs'],
        'name' => ['name', 'name2', 'name3'],
        'size' => ['size', 'size', 'size']
        # checksum, TimeDate not tripled
    };
    if (exists $map->{$field}) {
        for my $k (@{$map->{$field}}) {
            $szr->put($k, @val); # Kinda icky # KLUDGE errors?
        }
    } else {
        $szr->put($field, @val); # Kinda icky, again # KLUDGE errors?
    }
    return $szr->write();
}

sub append_file {
    my ($s, $filename, $data) = @_;
    my $szr = $s->{szr};
    my $prev_node = $s->{dir}->[-1];
    my $new_node_off = $prev_node->{offset} + $prev_node->{size};

    $szr->read('RD_ROOT_OBJ', 0); # KLUDGE errors
    my $rdsk_size = $szr->get('size');
    my $rdsk_max = $szr->get('maxsize');

    $szr->read('RD_NODE_OBJ', $new_node_off); # KLUDGE errors

    # Error if stuff is too big
    return undef if $rdsk_size + $s->{rd_node_size} + length($data) > $rdsk_max;

    # Write data before we write the node
    my $data_pos = $new_node_off + $s->{rd_node_size};
    $szr->{fh}->seek($data_pos) or return undef;
    $szr->{fh}->write($data) == length($data) or return undef;

    # Now write the new node
    $s->rd_node_put('self_ofs', $new_node_off);
    $s->rd_node_put('next_ofs', 0);
    $s->rd_node_put('name', $filename);
    $s->rd_node_put('size', length($data));
    $s->rd_node_put('checksum', 0);
    my @time = localtime;
    $time[5] += 1900; # Fix year
    $time[4] += 1; # Fix month
    my $ti = 0;
    for my $field (qw(second minute hour day month year)) {
        $s->rd_node_put('TimeDate', $field, $time[$ti++]);
    }
    $szr->write or return undef;

    # Update the previous last node to point to this one
    $s->set_rd_node($prev_node, 'next_ofs', $new_node_off);

    # Update dir
    my $date = sprintf "%.04d-%.02d-%.02d %.02d:%.02d:%.02d",
        $time[5], $time[4], $time[3], $time[2], $time[1], $time[0];
    push(@{$s->{dir}}, { 'name'=>$filename, 'date'=>$date,
        'rd_node_offset'=>$new_node_off, 'offset'=>$data_pos,
        'size'=>length($data)});

    # Update the root object
    $szr->read('RD_ROOT_OBJ', 0); # KLUDGE errors
    my $new_size = $szr->get('size') + $s->{rd_node_size} + length($data);
    $szr->put('size', $new_size);
    $szr->put('size2', $new_size);
    $szr->put('size3', $new_size);
    $szr->write or return undef;
    return 1;
}

sub delete_file {
    my ($s, $filename) = @_;
    my $szr = $s->{szr};
    return undef if @{$s->{dir}} < 2;
    my $last_node = $s->{dir}->[-1];
    my $prev_node = $s->{dir}->[-2];

    return undef if defined $filename && $last_node->{name} ne $filename;

    my $reclaimed = $last_node->{size} + $s->{rd_node_size};

    $s->set_rd_node($prev_node, 'next_ofs', 0);

    $szr->read('RD_ROOT_OBJ', 0); # KLUDGE errors
    my $new_size = $szr->get('size') - $reclaimed;
    $szr->put('size', $new_size);
    $szr->put('size2', $new_size);
    $szr->put('size3', $new_size);
    $szr->write or return undef;
    pop @{$s->{dir}};
    return 1;
}

sub sys_trunc { # delete all files after the system files
    my ($s) = @_;
    my $dir = $s->get_dir;
    for my $f (reverse @$dir) {
        last if $f->{name} =~ /^\\system\\/;
        $s->delete_file($f->{name});
    }
    return 1;
}

# File ops begin here

sub open {
    my ($s, $fname) = @_;
    for my $e (@{$s->{dir}}) {
        next unless $fname eq $e->{name};
        $s->{current} = $e;
        $s->{offset} = $e->{offset};
        $s->{fpos} = $e->{offset};
        $s->{size} = $e->{size};
        return $e;
    }
    return undef;
}

sub fh_open { # KLUDGE!!! open() and other stuff should eventually go away...
    my ($s, $fname, $mode) = @_;
    $mode = SoMat::File::O_RDONLY() unless defined $mode;
    for my $e (@{$s->{dir}}) {
        return SoMat::File::Offset->new($s->{filename}, $mode,
            $e->{offset}, $e->{size}) if $fname eq $e->{name};
    }
    return undef;
}

sub _seek {
    my ($s) = @_;
    return undef if $s->{fpos} != $s->{fh}->seek($s->{fpos}, 0);
    my $spos = $s->{fpos} - $s->{offset};
    return undef if $spos < 0 || $spos >= $s->{size};
    return $spos;
}

sub read {
    my ($s, $size) = @_;
    $size = $s->{size} unless defined $size;
    $s->{fpos} = $s->{fh}->seek($s->{fpos}, 0); # KLUDGE errors
    my $spos = $s->{fpos} - $s->{offset};
    return undef if $spos < 0 || $spos >= $s->{size};
    $size = $s->size - $spos if $size > $size - $spos;
    my $buf;
    my $got = $s->{fh}->read($buf, $size); # KLUDGE check errors
    $s->{fpos} = $s->{fh}->seek(0, 1); # KLUDGE errors
    return $buf;
}

sub close {
    my ($s) = @_;
    return 1;
}

sub write {
    my ($s, $data, $len, $off) = @_;
    return undef unless defined $s->_seek;
    return $s->{fh}->write($data, $len, $off);
}

sub seek {
    my ($s, $pos, $whence) = @_;
    my $fpos = $s->{fpos} - $s->{offset};
    $fpos = $pos if $whence == 0;
    $fpos += $pos if $whence == 1;
    $fpos = $s->{size} + $pos if $whence == 2;
    return undef if $fpos < 0 || $fpos > $s->{size};
    $s->{fpos} = $fpos + $s->{offset};
    return $fpos;
}

sub append_data {
    my ($s, $filename, $data) = @_;
    my $szr = $s->{szr};
    my $node = $s->get_rd_node($filename);
    return undef unless $node == $s->{dir}->[-1]; # KLUDGE must be last node

    $szr->{fh}->seek($node->{offset} + $node->{size}) or return undef;
    $szr->{fh}->write($data) == length($data) or return undef;
    $node->{size} += length($data);
    $s->set_rd_node($node, 'size', $node->{size}); # KLUDGE errors
    return 1;
}

1;
