
package SoMat::eDAQ::SerialFilter;

require SoMat::eDAQ::Hardware::Serial;
require SoMat::Network::Dispatch;
require SoMat::File;

use constant Debug => 0;

sub new {
    my $class = shift;
    my $s = bless {}, ref($class) || $class;
    $s->{inbuf} = "";
    $s->{textbuf} = "";
    return $s;
}

sub run {
    my ($s, $tty, $edaqdev, @args) = @_;

    my $dsp = SoMat::Network::Dispatch->new();
    $s->{dispatch} = $dsp;

print STDERR "Mucking pgetty\n" if Debug;
    if (defined (my $fh = SoMat::File->new(">/var/state/pgetty.conf"))) {
        print $fh "{ 'device' => '' };";
        close $fh;
    }
    kill 1, `cat /var/run/pgetty.pid`; # HUP
    select(undef, undef, undef, 0.5);

    my $ttyfh = SoMat::eDAQ::Hardware::Serial->open($tty, @args);
    die "Can't open $tty: $!" unless defined $ttyfh;

    $dsp->add('R', $ttyfh, $s->can('read_tty'), $s);

print STDERR "Opening $edaqdev\n" if Debug;
    my $edaqfh = SoMat::File->new($edaqdev, 0); # O_RDONLY
    die "Can't open $edaqdev: $!" unless defined $edaqfh;
    $dsp->add('R', $edaqfh, $s->can('read_edaqserial'), $s);

    $dsp->add('R', \*STDIN, $s->can('read_exit'), $s);

print STDERR "Calling dispatch\n" if Debug;
    $dsp->dispatch while 1;
}

sub read_exit {
    my ($fh, $s) = @_;
    my $nothing;
    exit 0 if sysread($fh, $nothing, 8192) <= 0;
}

sub read_tty {
    my ($fh, $s) = @_;
print STDERR "In read_tty\n" if Debug;
    my $nothing;
    sysread $fh, $nothing, 8192;
print STDERR "    GOT: $nothing\n" if Debug;
}

sub read_edaqserial {
    my ($fh, $s) = @_;
print STDERR "In read_edaqserial\n" if Debug;
    my $buf = $s->{inbuf};
    my $got = sysread $fh, $buf, 8192, length($buf);
print STDERR "    GOT $got bytes...\n" if Debug;
if (Debug) {
    my $hex = $buf;
    $hex =~ s/(.)/sprintf(" %02X", ord($1))/seg;
print STDERR "    $hex\n" if Debug;
}

    while (length($buf) >= 2) {
        my $buf_len = length($buf);
        my $header_len = unpack("v", $buf);
        my $packet_len = $header_len + ($header_len & 1) + 2;
print STDERR "packet_len = $packet_len, buf_len = $buf_len;\n" if Debug;
        last if $packet_len > length($buf);
        my $pkt = substr($buf, 0, $packet_len, '');
print STDERR "after substr, buf_len = ", length($buf), "\n" if Debug;
        my ($len, $port, $timestamp, $str) = unpack "vvVa*", $pkt;
print STDERR "len = $len, port = $port, timestamp = $timestamp\n" if Debug;
        $str = substr($str, 0, $header_len - 6);
print STDERR "str len = ", length($str), "\n" if Debug;
print STDERR "GOT str = $str\n" if Debug;
        $s->{textbuf} .= $str;
        while ($s->{textbuf} =~ s/^(.*\n)//) {
print STDERR "CALLING DECODE: $1\n" if Debug;
            $s->{timestamp} = $timestamp;
            $s->decode($1);
        }
        send_heartbeat($timestamp) unless length($str);
    }
#print STDERR "TEXTBUF: $s->{textbuf}\n";
    $s->{inbuf} = $buf;
}

sub send_heartbeat (\$) {
    my ($timestamp) = @_;
    my $pkt = pack "vvV", 6, 0, $timestamp;
    send_pkt($pkt);
}

sub send_pkt (\$) {
    syswrite STDOUT, $_[0];
}

1;

