use strict;

package SoMat::eDAQ::TCE::ParseTCE;

use vars qw($VERSION);
$VERSION = sprintf("%d.%02d", q$Revision: GIT $ =~ /(\d+)\.(\d+)/);

use Carp qw(cluck);

sub new {
    my $class = shift;
    my $s = {};
    bless $s, ref $class || $class;
    $s->{'args'} = (ref($_[0]) eq "HASH") ? $_[0] : { @_ };
    return $s;
}

sub read_tce {
    my $class = shift;
    my $filestring = shift;

    my $ini = [];
    my $cat = undef;
    my $entries = {};
    my $n = 0;
    for my $line (split /\n/, $filestring) {
        chomp $line;
        $line =~ s/\r$//;
        ++$n;
        if ($line =~ /^\[(\w+)\]\s*$/) {
            $cat = $1;
            $entries = {};
            push(@$ini, [ $cat, $entries]);
            next;
        }
        if ($line =~ /^(\w+)=(.*)$/) {
            $entries->{$1} = $2;
            next;
        }
        if ($line =~ /^\s*$/) {
            next;
        }
        warn("Syntax error on line $n : $line\n");
    }
    return $ini;
}

sub tceparse {
    my ($s, $tce) = @_;

    my @modules = ();
    my %modules_seen;

    foreach my $e (@$tce) {
        my $cat = $e->[0];
        my $dat = $e->[1];
        bless $dat, 'SoMat::eDAQ::TCE::Item';
        if ($cat !~ /^Hard/ && exists $dat->{'Prefix'}) {
            my $prefix = $dat->{'Prefix'};
            push(@modules, $prefix) unless exists $modules_seen{$prefix};
            $modules_seen{$prefix} = 1;
        }
    }
    $s->{'__Modules'} = \@modules;
    $s->{'tce'} = $tce;
    return $tce;
}

sub write {
    my $s = shift;
    print @_;
}

sub writeline {
    my $s = shift;
    print @_, "\r\n";
}

sub profset {
    my $s = shift;
    my $id = shift;
    return unless @_;
    $s->writeline("PROF_set \%$id ", join(" ", @_));
}

sub common_channel_stuff {
    my $s = shift;
    my ($id, $ch) = @_;
    $s->profset($id, 'Description', '"' . $ch->val('Description_1') . '"');
    $s->profset($id, 'Type', '"' . $ch->val('Type_1') . '"');
    $s->profset($id, 'Units', '"' . $ch->val('Units_1') . '"');
    #$s->profset($id, 'SampleRate', sprintf("%g", $ch->val('SampleRate')));
    $s->profset($id, 'SampleRate', $s->get_sample_rate($id, $ch));
    $s->profset($id, 'FSMin', sprintf("%g", $ch->val('FS_Min_1')));
    $s->profset($id, 'FSMax', sprintf("%g", $ch->val('FS_Max_1')));
}

sub get_sample_rate {
    my $s = shift;
    my ($id, $ch) = @_;
die("get_sample_rate failed, id = $id, undefined ch") unless defined $ch;
    my $rate = $ch->val('SampleRate');
    unless (defined $rate) {
        $rate = "XXX"; # Flag error if badness
        my $nchannels = $ch->val('InputChs');
        for (my $n = 0; $n < $nchannels; $n++) {
             my $chname = "ch_" .  $ch->val(sprintf("InputCh_%d", $n + 1));
             my $cch = $s->{'channels'}->{$chname};
             # KLUDGE should probably do GCD thing here????
             #$rate = $cch->val('SampleRate') if $cch;
             my $nr = $s->get_sample_rate($chname, $cch);
             $rate = $nr if $nr > 0;
        }
    }
    my $factor = 1;
    my $prefix = $ch->val('Prefix');
    if ($prefix eq 'DSAMPLE') {
        $factor /= $ch->val('Factor');
    } elsif ($prefix eq 'USAMPLE') {
        $factor *= $ch->val('Factor');
    }
    $rate *= $factor;
    return sprintf("%g", $rate);
}

sub write_input_channels {
    my $s = shift;
    my ($id, $dm) = @_;
    my $nchannels = $dm->val('InputChs');
    $s->profset($id, 'NumInputChs', $nchannels);
    for (my $n = 0; $n < $nchannels; $n++) {
        # KLUDGE onode.0???
        $s->profset($id, "inode.$n",
            sprintf("(PROF_get %%ch_%s onode.0)",
                $dm->val(sprintf("InputCh_%d", $n + 1))));
    }
}

sub tce2bat {
    my $s = shift;
    my $tce = shift;
    my $i;

    $s->writeline();
    $s->writeline("SHL_rem *** Engine ***");
    $s->writeline("SHL_set engine.0 (PROF_create)"); # KLUDGE???

    $s->writeline();
    $s->writeline("SHL_rem *** Load Modules ***");
    foreach my $mod (@{$s->{'__Modules'}}) {
        $s->writeline("MODULE_load $mod ", lc($mod));
    }

    $s->{'channels'} = {};
    $i = 1;
    foreach my $item (@$tce) {
        next unless $item->[0] =~ /^ChanItem_/;
        my $ch = $item->[1];
        my $id = "ch_" . $ch->val('ID_1');
        $s->{'channels'}->{$id} = $ch;
        my $prefix = $ch->val('Prefix');
        $s->writeline();
        $s->writeline("SHL_rem *** Chanmode ***");
        $s->writeline("SHL_set $id (PROF_create)");
        $s->profset($id, 'NumIDs', 1);
        $s->profset($id, 'ID.0', $ch->val('ID_1'));
        (my $connector = $ch->val('Connector')) =~ s/\.\w+$//;
        $s->profset($id, 'Connector', $connector);
        $s->common_channel_stuff($id, $ch);
        $s->profset($id, 'NumInputFrames', 64);
        $s->profset($id, 'CalSlope', sprintf("%g", $ch->val('CalSlope')));
        $s->profset($id, 'CalIntercept', sprintf("%g", $ch->val('CalIntercept')));
        (my $devid) = $id =~ /(\d+)$/;
        $devid = 1 unless defined $devid;
        if ($prefix eq 'XDMS_LOLEV') {
            --$devid;
            $devid ^= 0x04;
        } else {
            --$devid;
        }
        $s->profset($id, 'DeviceID', $devid); # KLUDGE???
        $s->profset($id, 'onode.0', '(LINX_create)');
        $s->profset($id, 'odtype.0', $ch->val('ChanDataType'));
        my $iid = $i - 1;
        $s->profset('engine.0', "Dev${iid}Init (MODULE_address $prefix)");
        $s->profset('engine.0', "Dev${iid}Profile \%$id");
        ++$i;
    }
    my $nchannels = $i - 1;

# [SoftItem_1]
# Prefix=USAMPLE
# NumIDs=1 
# ID_1=us_switch
# Description_1=Digital BWI
# Type_1=Logical
# Units_1=logic
# ChanDataType=264
# FS_Min_1=0
# FS_Max_1=0
# InputChs=1
# InputCh_1=switch
# Factor=50
# ObjectID=0
# 
# SHL_rem *** Softmode ***
# SHL_set ch_us_switch (PROF_create)
# PROF_set %ch_us_switch NumIDs 1
# PROF_set %ch_us_switch ID.0 us_switch
# PROF_set %ch_us_switch Description "Digital BWI"
# PROF_set %ch_us_switch Type "Logical"
# PROF_set %ch_us_switch Units "logic"
# PROF_set %ch_us_switch SampleRate 2500
# PROF_set %ch_us_switch FSMin 0
# PROF_set %ch_us_switch FSMax 0
# PROF_set %ch_us_switch NumInputChs 1
# PROF_set %ch_us_switch inode.0 (PROF_get %ch_switch onode.0)
# PROF_set %ch_us_switch Factor 50
# PROF_set %ch_us_switch onode.0 (LINX_create)
# PROF_set %ch_us_switch odtype.0 264
# PROF_set %engine.0 Mod0Init (MODULE_address USAMPLE)
# PROF_set %engine.0 Mod0Profile %ch_us_switch

    my $nengines = 1; # KLUDGE!!!
    my $engines = {};

    my $nmodules = 0;
    $i = 1;
    foreach my $item (@$tce) {
        next unless $item->[0] =~ /^SoftItem_/;
        my $ch = $item->[1];
        my $id = "ch_" . $ch->val('ID_1');
        $s->{'channels'}->{$id} = $ch;
        my $prefix = $ch->val('Prefix');
        $s->writeline();
        $s->writeline("SHL_rem *** Softmode ***");
        $s->writeline("SHL_set $id (PROF_create)");
        $s->profset($id, 'NumIDs', 1);
        $s->profset($id, 'ID.0', $ch->val('ID_1'));
        #(my $connector = $ch->val('Connector')) =~ s/\.\w+$//;
        #$s->profset($id, 'Connector', $connector);
        $s->common_channel_stuff($id, $ch);
        $s->write_input_channels($id, $ch);
        if ($ch->val('Factor')) { # KLUDGE!!!
            $s->profset($id, 'Factor', $ch->val('Factor'));
        }
        $s->profset($id, 'onode.0', '(LINX_create)');
        $s->profset($id, 'odtype.0', $ch->val('ChanDataType'));
        my $iid = $i - 1;
        $s->profset('engine.0', "Mod${iid}Init (MODULE_address $prefix)");
        $s->profset('engine.0', "Mod${iid}Profile \%$id");
        ++$i;
        ++$nmodules;
    }

    foreach my $item (@$tce) {
        next unless $item->[0] =~ /^DataItem_/;
        my $dm = $item->[1];
        my $id = "dm_" . $dm->val('ID');
        $s->writeline();
        $s->writeline("SHL_rem *** Datamode ***");
        $s->writeline("SHL_set $id (PROF_create)");
        $s->profset($id, 'ID', $dm->val('ID'));

if (1) {
        my $nchannels = $dm->val('InputChs');
        my $samplerate = "XXX"; # Flag error if badness
        for (my $n = 0; $n < $nchannels; $n++) {
             my $chname = "ch_" .  $dm->val(sprintf("InputCh_%d", $n + 1));
             my $cch = $s->{'channels'}->{$chname};
             # KLUDGE should probably do GCD thing here????
             $samplerate = $cch->val('SampleRate') if $cch;
        }
        $s->profset($id, 'SampleRate', $samplerate);
        $s->profset($id, 'NumInputChs', $nchannels);
        for (my $n = 0; $n < $nchannels; $n++) {
            # KLUDGE onode.0???
            $s->profset($id, "inode.$n",
                sprintf("(PROF_get %%ch_%s onode.0)",
                    $dm->val(sprintf("InputCh_%d", $n + 1))));
        }
} else {
        $s->profset($id, 'SampleRate', $s->get_sample_rate($id, $dm));
        $s->write_input_channels($id, $dm);
}
        $s->profset($id, 'DataType', $dm->val('DataType'));
        $s->profset($id, 'HaveTimeChan', $dm->val('HaveTimeChan'));
        $s->profset($id, 'PCMStorage', $dm->val('PCMStorage'));
        my $trigmap = {
            'Always On', 0
        };
        # KLUDGE check condition here
        $s->profset($id, 'TriggerMode', $trigmap->{$dm->val('TrigCond')});
        $s->profset($id, 'TrackMinMax', 0); # KLUDGE!!!
        $s->profset($id, 'Dim1Label', "Time");
        $s->profset($id, 'Dim1Units', "secs");
        $s->profset($id, 'LabelDesc', $dm->val('ObjectID'));

        my $eid = "engine.0";   # KLUDGE
        $s->profset($eid, "Mod${nmodules}Init",
            "(MODULE_address " . $dm->val('Prefix') . ")");
        $s->profset($eid, "Mod${nmodules}Profile", "\%$id");
        ++$nmodules;
    }

    $s->writeline();
    $s->writeline("SHL_rem *** Execute ***");
    for ($i = 0; $i < $nengines; $i++) {
        $s->profset("engine.$i", "NumDevices", $nchannels);
        $s->profset("engine.$i", "NumModules", $nmodules);
        $s->writeline("LINX_setup \%engine.$i");
    }
    #$s->writeline();
}

package SoMat::eDAQ::TCE::Item;

sub new {
    my $class = shift;
    my $s = {};
    bless $s, ref $class || $class;
    $s->{'__args'} = (ref($_[0]) eq "HASH") ? $_[0] : { @_ };
    return $s;
}

sub val {
    my ($s, $name) = @_;
    return undef unless exists $s->{$name};
    return $s->{$name};
}

1;
