package SoMat::eDAQ::Web::Commands;

use strict;
use warnings 'all';

require SoMat::TCE::File;
require SoMat::eDAQ::SRam;

my $re_iso_8601 = qr{^(\d{4})-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)(\.\d+)?([\+\-]\d+)?$};

my $SIE_Dir = '/hd/sie';

my $statics = {
    'jsontest/clock'=>sub {
        my ($s, $q, $p) = @_;
        $s->suffixes('json');
        #$s->send_header('text/plain', 200, 'OK'); # KLUDGE JSON => Javascript
        my $clock = $s->sys('/bin/date', '-Isec'); # KLUDGE
        $clock =~ s/\s+$//s;
        print qq({ "clock": "$clock" }\n);
    },
    'json/status'=>sub {
        my ($s, $q, $p) = @_;
        $s->suffixes('json');

        require SoMat::Network::WebRequest;
        my $url = "http://127.0.0.1:2345/teststatus";
        my $c = {LocalTime=>scalar(localtime)};
        eval {
            my $web = SoMat::Network::WebRequest->new($url);
            while (defined (my $line = $web->getline)) {
                $c->{$1} = $2 if $line =~ /^(\w+)=([^\r\n]*)/;
            }
            if (exists $c->{SetupFileLong}) {
                ($c->{SetupName} = $c->{SetupFileLong}) =~ s/\.tce$//i;
            }
        };

        my $v;
        print "{\n";
        print join(",\n", map {
            (($v = $c->{$_}) =~ s/([\'\\])/\\$1/, qq{'$_' : '$v'})[-1]
        } keys %$c);
        print "\n}\n";
    },
    'debug'=>sub {
        my ($s, $q, $p) = @_;
        print "debug, METHOD=$s->{state}{HTTP_Method}\n";
        my $hdrs = $s->{state}{HTTP_Headers};
        for my $k (sort keys %$hdrs) {
            print "Header: $k => $hdrs->{$k}\n";
        }
        for my $k (sort keys %$q) {
            print "Q: $k => $q->{$k}\n";
        }
        # KLUDGE!!!
        #my $content = $s->{web}->read_post($s->{state}{HTTP_Headers});;
        my $content = $s->{state}{HTTP_Content};
        $content = '' unless defined $content;
        print "\n";
        printf "CONTENT: %d bytes\n", length($content);
        print $content;
    },
    'debug/echo'=>sub {
        my ($s, $q, $p) = @_;
        print "-/debug/echo, METHOD=$s->{state}{HTTP_Method}\n";
        my $hdrs = $s->{state}{HTTP_Headers};
        for my $k (sort keys %$hdrs) {
            print "HEADER: $k => $hdrs->{$k}\n";
        }
        my $th = $s->{state}{HTTP_Query_Hash};
        for my $k (sort keys %$th) {
            for my $hf (@{$th->{$k}}) {
                printf " QUERY: $k => $hf\n";
            }
        }
        $th = $s->{state}{HTTP_Post_Hash};
        for my $k (sort keys %$th) {
            for my $hf (@{$th->{$k}}) {
                if (ref($hf) eq 'HASH') {
                    print "  POST: $k => {";
                    print join(', ', map { "$_=>'$hf->{$_}'" } sort(keys %$hf));
                    print "}\n";
                    next;
                }
                printf "  POST: $k => $hf\n";
            }
        }
    },
    'system/clock'=>sub {
        my ($s, $q, $p) = @_;
        if ($q) {
            $s->rw;
            my $utc = $q->{utc} || '';
            my @d = $utc =~ /$re_iso_8601/
                or die "'utc' ($utc) must be in 'YYYY-MM-DDThh:mm:ss' format";
            my $f = {};
            for my $field (qw(year month day hour minute second fraction tz)) {
                $f->{$field} = shift @d;
            }
            my $dateutc = sprintf '%02d%02d%02d%02d%04d.%02d',
                                 #  MM  DD  hh mm  YY   ss
                $f->{month}, $f->{day}, $f->{hour}, $f->{minute},
                $f->{year}, $f->{second};
            $s->sys('/bin/date', '-u', '-s', $dateutc);
            print "System time is now: ", $s->sys('/bin/date', '-Iseconds');
            $s->sys('/sbin/hwclock', '--systohc', '--utc');
        } else {
            print $s->sys('/bin/date', '-Iseconds');
        }
    },
    'system/file/test'=>sub {
        my ($s, $q, $p) = @_;
        $s->send_header('text/html', 200, 'OK');
        print <<'EOT';
<html>
<head>
<title>system/file/test</title>
</head>
<body>
<h1>system/file/test</h1>
<ul>

<h2>GET</h2>
<form method="GET" action="/~/debug">
<textarea name="payload" rows="6" cols="60">GET / normal</textarea><br>
<input type="submit" name="__commit" value="Transfer">
</form>

<h2>POST</h2>
<form method="POST" action="/~/debug">
<textarea name="payload" rows="6" cols="60">POST / normal</textarea><br>
<input type="submit" name="__commit" value="Transfer">
</form>

<h2>POST file</h2>
<form method="POST" action="/~/debug" enctype="multipart/form-data">
<input type="file" name="payload" size="60" value="foo"><br>
<input type="submit" name="__commit" value="Transfer">
</form>

</body>
</html>
EOT
    },
    'system/file/find'=>sub {
        my ($s, $q, $p) = @_;
#print STDERR "system/file/put, METHOD=$s->{state}{HTTP_Method}\n";
        my $dir = $q->{dir};
        my $pat = $q->{pattern};
        if (length($pat)) {
            $pat = qr{$pat};
        } else {
            $pat = undef;
        }
        my $files = $s->file_find($dir, $pat);
        for my $f (@$files) {
            print "$f\n"; # KLUDGE
        }
    },
    'system/file/put'=>sub {
        my ($s, $q, $p) = @_;
#print STDERR "system/file/put, METHOD=$s->{state}{HTTP_Method}\n";
        $s->rw;
        # KLUDGE fix read_post, create infrastructure...
        my $content = $s->{web}->read_post($s->{state}{HTTP_Headers});;
        $content = '' unless defined $content;
        my $fld = 'file'; # KLUDGE destination/sink/output/???
        my $fname = $q->{$fld};
        open(my $fh, '>', $fname) or die "Can't open '$fld' ($fname): $!";
        print $fh $content;
        close($fh) or die "Error closing '$fld' ($fname): $!";
        printf "Wrote '$fld' ($fname) successfully: %d bytes\n", -s $fname;
    },
    'system/file/list'=>sub {
        my ($s, $q, $p) = @_;
        $s->suffixes('json', 'perl', 'tce');
#print STDERR "system/file/put, METHOD=$s->{state}{HTTP_Method}\n";
        my $dir = $q->{dir};
        $dir =~ m(^/) or die "dir argument must be an absolute directory";
        my $pat = $q->{pattern};
        if (length($pat)) {
            $pat = qr{$pat};
        } else {
            $pat = undef;
        }
        my $files = $s->file_find($dir, $pat);
        my $i = 0;
        for my $f (@$files) {
            my $r = $s->file_stat($f);
            if (-d "$f") {
                $r->{type} = 'directory';
            } else {
                my $esc_f = $s->{web}->escape_url($f);
                $r->{get_url} = "/-/system/file/get.bin?file=$esc_f";
                $r->{delete_url} = "/~/system/file/delete.bin?file=$esc_f";
                $r->{md5} = get_checksum($f, 'md5') if $q->{md5};
                $r->{sha1} = get_checksum($f, 'sha1') if $q->{sha1};
            }
            #print "$f\n"; # KLUDGE
            $s->emit_dispatch($r, "file_$i", $s->{suffix});
            ++$i;
        }
    },
    'system/file/get'=>sub {
        my ($s, $q, $p) = @_;
        my $path = $q->{file};
        my $stat = $s->file_stat($path);
        die "No such file: $path" unless $stat;
        (my $name = $path) =~ s(.*/)();
        $s->unbuffer_stdout;
        my $heads = qq(Content-Disposition: attachment; filename="$name"\r\n);
        $heads .= "Content-Length: $stat->{size}\r\n";
        # KLUDGE application/x-binary because Microsoft does a totally stupid
        # attempt at filetype autodetection with application/octet-stream!
        $s->send_header('application/x-binary', 200, "OK", $heads);

        open(my $fh, '<', "$path") or return 1;
        my $data;
        while (sysread($fh, $data, 65536)) {
            print $data or return 1;
        }
        close($fh);
    },
    'system/file/delete'=>sub {
        my ($s, $q, $p) = @_;
        $s->rw;
        my $path = $q->{file};
        my $stat = $s->file_stat($path);
        die "No such file: $path" unless $stat;
        unlink($stat->{name}) or die "Couldn't delete file: $!";
        print "OK\n";
    },
    'tce/init'=>sub {
        my ($s, $q, $p) = @_;
        my $ok = 1;
        $s->rw;
        $s->suffixes('tce');
        require SoMat::eDAQ::Hardware::List;
        my $serial_number = SoMat::eDAQ::Hardware::List->get_mpb_serial_number();
        if ($q->{tce}) {
            my $messages = "";
            my $auxdir = "/hd/tce_aux_files";
            my $setup_name = $q->{setup_name} || "test";
            require SoMat::TCE::File;
            require SoMat::Network::WebRequest;
            my $chs = SoMat::TCE::File->new->read_string($q->{tce});
            $chs->remap_aux_files($auxdir);

            if ($chs->get_network_nodes > 1) {

                my $my_chs = SoMat::TCE::File->new->read_string($q->{tce});
                $my_chs->make_my_tce_file($serial_number);

                # Get Node Information
                my $nodename = $my_chs->get_nodename_from_serial($serial_number);
                $nodename or die "NodeName for serial number '$serial_number' returned undef.\n";
                my $nodetype = $my_chs->get_nodetype_from_serial($serial_number);
                $nodetype or die "NodeType for serial number '$serial_number' returned undef.\n";

                unless ($nodetype == 2) {
                    require SoMat::eDAQ::Discovery;
                    my $edaqs = SoMat::eDAQ::Discovery->discover_edaqs;
                    # $edaqs is now a hash of MPB serial numbers -> ip addresses.
                    for my $slave (@{$chs->get_slaves}) {
                        my $serial = $chs->get_serial_from_nodename($slave);
                        # Forward tce to Slaves.
                        my $url = "http://$edaqs->{$serial}/~/tce/init.tce";
                        my $args = {
                            tce=>$chs->write_string,
                            setup_name=>$setup_name
                        };

                        my $web = SoMat::Network::WebRequest->new(url=>$url, post_form_data=>$args);

                        my $rc = $web->get_response_code;
                        my $rm = $web->get_response_message;
                        $ok = $ok && $rc == 200;

                        my $line ="";
                        while (defined ($line = $web->getline)) {
                            $messages .= $line;
                        }
                    }
                } 
                $chs->write_file('/tmp/proxy_tce_setup.tce');
                $my_chs->write_file("/tmp/$setup_name.tce");
            } else {
                $chs->write_file("/tmp/$setup_name.tce");
            }
            my $url = "http://127.0.0.1/~/tce/init_node.tce";
            my $args = {
                setup_name=>$setup_name
            };
            my $web = SoMat::Network::WebRequest->new(url=>$url, post_form_data=>$args);
            my $rc = $web->get_response_code;
            my $rm = $web->get_response_message;
            $ok = $ok && $rc == 200;
            my $line = "";
            while (defined ($line = $web->getline)) {
                $messages .= $line;
            }

            my $message_count = 0;
            my $new_messages = "";
            for $line (split /\n/,$messages) {
                if ($line =~ s/^(\[message_)(\d)\]/$1$message_count\]/) {
                    ++$message_count;
                }
                $new_messages .= "$line\n"; 
            }
            print $new_messages;

            die "reply with a 500 please" unless $ok;
        }
    },
    'tce/aom_file'=>sub {
        my ($s, $q, $p) = @_;
        my $ok = 1;
        $s->suffixes('aom','aox');
        my $status = $s->{web}->test_status;
        die "Test must be running to download AOM file."
            unless ($status->{TestInitialized} &&
                    ($status->{RunStarted} || $status->{RunPreview}));
        my $chs = SoMat::TCE::File->new->read_current();
        require SoMat::eDAQ::Hardware::List;
        my $serial_number =
            SoMat::eDAQ::Hardware::List->get_mpb_serial_number();
        require SoMat::TCE::File;
        require SoMat::Network::WebRequest;
        my $info = {};
        my $edaqs = {};
        my @nodes = $chs->get_network_nodes;
        if (@nodes > 1) {
            require SoMat::eDAQ::Discovery;
            $edaqs = SoMat::eDAQ::Discovery->discover_edaqs;
        }
        $edaqs->{$serial_number} = '127.0.0.1';
        # $edaqs is now a hash of MPB serial numbers -> ip addresses.

        for my $node (@nodes) {
            my $serial = $chs->get_serial_from_nodename($node);
            my $ip = $edaqs->{$serial};
            my $url = "http://$ip/-/internal/analog_output.txt";
            my $web = SoMat::Network::WebRequest->new(url=>$url);

            my $rc = $web->get_response_code;
            my $rm = $web->get_response_message;
            $ok = $ok && $rc == 200;

            my $o = "";
            my $line;
            while (defined ($line = $web->getline)) {
                chomp $line;
                $info->{$node}{$1} = $2 if $line =~ m(^([^=]+)=(.*)$);
            }
        }

        die "reply with a 500 please" unless $ok;

        my $o = SoMat::TCE::File->new;
        $o->set('Main', 'SetupFile', $status->{SetupFileLong});
        $o->set('Main', 'RunNumber', $status->{Run});

        my $next = 1;
        for my $sec ($chs->sections) {
            next if $sec !~ m(^ChanItem);
            my $node = $chs->get($sec, 'NodeName');
            my $id = $chs->get($sec, 'ID_1');
            my $conn = $chs->get($sec, 'Connector');
            next unless defined $info->{$node}{$conn};
            my ($scale, $offset, $postgain) =
                split(',', $info->{$node}{$conn});
            $scale *= $postgain if defined $postgain;
            my $os = "XdcrCh_$next";
            $o->set($os, 'ID', $id);
            $o->set($os, 'Units', $chs->get($sec, 'Units_1'));
            $o->set($os, 'NodeName', $node);
            $o->set($os, 'Connector', $conn);
            my $fsmin = $chs->get($sec, 'FS_Min_1');
            my $fsmax = $chs->get($sec, 'FS_Max_1');
            if ($s->{suffix} eq 'aox') { # KLUDGE ORDERING
                $o->set($os, 'Eng_FSMin', sprintf("%g", $fsmin));
                $o->set($os, 'Eng_FSMax', sprintf("%g", $fsmax));
            }
            $o->set($os, 'Scale', sprintf("%g", $scale));
            $o->set($os, 'Offset', sprintf("%g", $offset));
            if ($s->{suffix} eq 'aom') { # KLUDGE ORDERING
                $o->set($os, 'Eng_FSMin', sprintf("%g", $fsmin));
                $o->set($os, 'Eng_FSMax', sprintf("%g", $fsmax));
            }
            my $vmin = 0;
            my $vmax = 0;
            unless ($scale == 0) {
                $vmin = ($fsmin - $offset) / $scale;
                $vmax = ($fsmax - $offset) / $scale;
            }
            $o->set($os, 'Volts_FSMin', sprintf("%g", $vmin));
            $o->set($os, 'Volts_FSMax', sprintf("%g", $vmax));
            $o->set('Main', 'NumXdcrChs', $next);
            ++$next;
        }

        if ($s->{suffix} eq 'aom') {
            $o->write(*STDOUT, "\r\n");
        } else {
            for my $k ($o->keys('Main')) {
                my $v = $o->get('Main', $k);
                print "$k=$v\r\n";
            }
            print "\r\n";
            my @keys;
            for my $k ($o->keys('XdcrCh_1')) {
                push(@keys, $k);
            }
            print(join("\t", @keys, ''), "\r\n"); # KLUDGE trailing tab
            for my $sec ($o->sections) {
                next if $sec !~ m(^XdcrCh);
                my @line;
                for my $k (@keys) {
                    push(@line, $o->get($sec, $k));
                }
                print(join("\t", @line, ''), "\r\n"); # KLUDGE trailing tab
            }
        }
    },
    'test/data/sie/list'=>sub {
        my ($s, $q, $p) = @_;
        $s->suffixes('json', 'perl', 'tce');

        my $r = $s->sie_file_list;
        $s->emit_dispatch($r, 'DataFile', $s->{suffix});
    },
};

my $regexps = [
    qr{^test/(.*?)/data/get/(.*)$}=>sub {
        my ($s, $q, $p, $matches) = @_;

        $s->suffixes('sie');
        my $setup = $matches->[0];
        my $siename = $matches->[1];

        my $stat = $s->sie_file_stat($setup);
        die "No such SIE file: $setup" unless $stat;

        my $dir = $stat->{directory};
        $siename = $s->{web}->unescape_url($siename);
        $siename = $stat->{name} if $siename eq 'file';

#$s->send_header('text/plain', 200, "OK");

        opendir(my $dh, $dir) or die "Can't open $dir: $!";
        my @files = grep { /^sie_\d+\.si$/ } readdir($dh);
        closedir($dh);

        $s->unbuffer_stdout;
        $s->send_sie_header("$siename.sie", $stat->{size});

        for my $f (sort @files) {
            open(my $fh, '<', "$dir/$f") or return 1;
            my $data;
            while (sysread($fh, $data, 65536)) {
                print $data or return 1;
            }
            close($fh);
        }
    },
    qr{^test/(.*?)/data/delete$}=>sub {
        my ($s, $q, $p, $matches) = @_;

        $s->suffixes('txt');
        my $setup = $matches->[0];

        my $stat = $s->sie_file_stat($setup);
        die "No such SIE file: $setup" unless $stat;
        my $status = $s->{web}->test_status;
        die "Can't delete initialized SIE file"
            if $status->{TestInitialized} && $status->{ID} eq $stat->{id};

        $s->sys('/bin/rm', '-rf', $stat->{directory});
        print "OK\n";
    },
    qr{^test/(.*?)/end$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        $s->suffixes('txt');
        my $setup = $matches->[0];
        die "Only _DEFAULT_ test identifier is currently supported: $setup" unless $setup eq '_DEFAULT_';

        my $ok = 1;
        # Evaluate Network Status
        my $tce_file = SoMat::TCE::File->new->read_current();
        die "Cannot end if test is not initialized." unless $s->{web}->test_status->{TestInitialized};
         
        if ($tce_file->get_network_nodes > 1) {
            require SoMat::eDAQ::Discovery;
            require SoMat::eDAQ::Hardware::List;
            my $serial_number = SoMat::eDAQ::Hardware::List->get_mpb_serial_number;
            # Get Node Information
            my $nodetype = $tce_file->get_nodetype_from_serial($serial_number);
            $nodetype or die "NodeType for serial number '$serial_number' returned undef.\n";

            unless ($nodetype == 2) {
                my $edaqs = SoMat::eDAQ::Discovery->discover_edaqs;
                for my $slave (@{$tce_file->get_slaves}) {
                    my $serial = $tce_file->get_serial_from_nodename($slave);
                    # Forward tce to Slaves.
                    my $url = "http://$edaqs->{$serial}/~/test/$setup/end_node.txt";
                    my $web = SoMat::Network::WebRequest->new($url);
                    my $rc = $web->get_response_code;
                    my $rm = $web->get_response_message;

                    print "Slave $slave $edaqs->{$serial} Response: $rc $rm\n";
                    my $line ="" ;
                    $ok = $ok && $rc == 200;
                    while (defined (my $line = $web->getline)) {
                        print $line;
                    }
                }
                my $url = "http://127.0.0.1/-/test/$setup/end_node.txt\n";
                my $web = SoMat::Network::WebRequest->new($url);
                my $rc = $web->get_response_code;
                my $rm = $web->get_response_message;

                print "My Response: $rc $rm\n";
                my $line="";
                $ok = $ok && $rc == 200;
                while (defined (my $line = $web->getline)) {
                    print $line;
                }
            } else {
               warn "I appear to be a slave '$nodetype'\n";
               print "This eDAQ is configured as a slave in a network configuration.  The 'end' sequence is handled by the master.\n";
            }
        } else {
            #Singleton
            my $url = "http://127.0.0.1/-/test/$setup/end_node.tce";
            my $web = SoMat::Network::WebRequest->new($url);
            my $rc = $web->get_response_code;
            my $rm = $web->get_response_message;

            print "My Response: $rc $rm\n";
            my $line="";
            $ok = $ok && $rc == 200;
            while (defined (my $line = $web->getline)) {
                print $line;
            }
        }
        die "reply with a 500 please" unless $ok;
    },
    qr{^test/(.*?)/data/sie/contents$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        $s->suffixes('json', 'perl', 'tce');
        my $setup = $matches->[0];
        my $r = $s->sie_contents($setup);
        $s->emit_dispatch($r, 'sie_contents', $s->{suffix});
    },
    qr{^test/(.*?)/data/sie/extract$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        $s->suffixes('sie');
        my $setup = $matches->[0];
        $s->sie_extract($setup, $q);
    },
    qr{^test/(?:(.*?)/)?initialize$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        my $setup = $matches->[0];
        $setup = '_DEFAULT_' unless defined $setup;
        print "regexps text/XXX/initialize, setup=$setup\n",
        $s->do_perl('/eDAQ/Test/InitTCE.pl');
    },
    qr{^test/(.*?)/start(.*)$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        my $ok = 1;
        $s->suffixes('txt');
        my $setup = $matches->[0];
        die "Only _DEFAULT_ test identifier is currently supported: $setup" unless $setup eq '_DEFAULT_';

        # Evaluate Network Status
        my $tce_file = SoMat::TCE::File->new->read_current();
        die "Cannot start if test is not initialized." unless $s->{web}->test_status->{TestInitialized};
         
        if ($tce_file->get_network_nodes > 1) {
            require SoMat::eDAQ::Discovery;
            require SoMat::eDAQ::Hardware::List;
            my $serial_number = SoMat::eDAQ::Hardware::List->get_mpb_serial_number;
            # Get Node Information
            my $nodetype = $tce_file->get_nodetype_from_serial($serial_number);
            $nodetype or die "NodeType for serial number '$serial_number' returned undef.\n";

            unless ($nodetype == 2) {
                my $edaqs = SoMat::eDAQ::Discovery->discover_edaqs;
                for my $slave (@{$tce_file->get_slaves}) {
                    my $serial = $tce_file->get_serial_from_nodename($slave);
                    # Forward tce to Slaves.
                    my $url = "http://$edaqs->{$serial}/~/test/$setup/start_node.txt?$s->{web}{state}{HTTP_Query}";
                    my $web = SoMat::Network::WebRequest->new($url);
                    my $rc = $web->get_response_code;
                    my $rm = $web->get_response_message;
                    $ok = $ok && $rc == 200;

                    print "Slave $slave $edaqs->{$serial} Response: $rc $rm\n";
                    while (defined (my $line = $web->getline)) {
                        print $line;
                    }
                }
                my $url = "http://127.0.0.1/-/test/$setup/start_node.txt?$s->{web}{state}{HTTP_Query}";
                my $web = SoMat::Network::WebRequest->new($url);
                my $rc = $web->get_response_code;
                my $rm = $web->get_response_message;
                $ok = $ok && $rc == 200;

                print "My Response: $rc $rm\n";
                while (defined (my $line = $web->getline)) {
                    print $line;
                }
            } else {
               warn "I appear to be a slave '$nodetype'\n";
               print "This eDAQ is configured as a slave in a network configuration.  The 'start' sequence is handled by the master.\n";
            }
        } else {
            #Singleton
            my $url = "http://127.0.0.1/-/test/$setup/start_node.tce?$s->{web}{state}{HTTP_Query}";
            my $web = SoMat::Network::WebRequest->new($url);
            my $rc = $web->get_response_code;
            my $rm = $web->get_response_message;

            print "My Response: $rc $rm\n";
            my $line="";
            $ok = $ok && $rc == 200;
            while (defined (my $line = $web->getline)) {
                print $line;
            }
        }
        die "reply with a 500 please" unless $ok;
    },
    qr{^retest(.*)$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        print "regexps retest, arg=$matches->[0]\n";
    },
    qr{^tce/(.*)$}=>sub {
        my ($s, $q, $p, $matches) = @_;
        die "'.tce' suffix only" unless $s->{suffix} eq 'tce';
        $s->rw;
        $q = $s->{state}{HTTP_Query};
        $q = "?$q" if defined $q;
        $q = '' unless defined $q;
        require SoMat::Network::WebRequest;
        my $url = "http://127.0.0.1:7733/tce/$matches->[0]$q";
        my $out = '';

        # KLUDGE Dunno if we need to eval() this or not...
        my $web = SoMat::Network::WebRequest->new($url);
        while (defined (my $line = $web->getline)) {
            $out .= $line;
        }
        print $out;
    },
];

sub new {
    my ($class, @args) = @_;
    my $s = bless { @args }, ref($class) || $class;
    $s->{state} = $s->{web}{state} if $s->{web};
    return $s;
}

sub run {
    my ($class, $web) = @_;

    my $s = bless {}, ref($class) || $class;

    $s->{state} = $web->{state};
    $s->{web} = $web;

    my $q = $s->{state}{HTTP_Query_Hash};
    if (keys %$q) {
        for my $k (keys %$q) {
            my $v = $q->{$k};
            $s->{query}{$k} = (@$v > 1) ? $v : $v->[0];
        }
    } else {
        $s->{query} = undef;
    }

    if ($s->{state}{HTTP_Method} eq 'POST') {
        if ($web->is_multipart) {
            $web->read_multipart;
            # KLUDGE now what?
            my $pq = $s->{state}{HTTP_Post_Hash};
            for my $k (keys %$pq) {
                my $t = $pq->{$k}[0];
                if (ref($t) eq 'HASH') {
                    # KLUDGES?
                    $s->{query}{"${k}.filename"} = $t->{filename};
                    #$s->{query}{"${k}.info"} = $t->{info};
                    ## HIDEOUS KLUDGE
                    local $/ = undef;
                    open(my $tfh, '<', $t->{file});
                    $t = <$tfh>;
                    close($tfh);
                }
                $s->{query}{$k} = $t;
            }
        } else {
            $s->{state}{HTTP_Content}
                = $web->read_post($s->{state}{HTTP_Headers});
            my $pq = $web->decode_query($s->{state}{HTTP_Content});
            for my $k (keys %$pq) {
                $s->{query}{$k} = $pq->{$k}[0];
            }
        }
    } elsif ($s->{state}{HTTP_Method} eq 'PUT') {
        $s->{state}{HTTP_Content}
            = $web->read_post($s->{state}{HTTP_Headers});
        $s->{query}{payload} = $s->{state}{HTTP_Content}
    }

    my $path = $s->{state}->{HTTP_Path};
    my @p = split(/\//, $path);

    $s->error if @p < 3 || $p[0] ne '' || $p[1] !~ /^[\-\~]$/;

    $s->{readonly} = ($p[1] eq '-');

    if ($p[-1] =~ s/\.(\w+)$//) {
        $s->{suffix} = $1;
    } else {
        $s->{suffix} = 'txt';
    }

    splice(@p, 0, 2);

    $s->{path} = \@p;

    my $request = join('/', @p);
    my $sub;
    my @matches;

    unless ($sub = $statics->{$request}) {
        my $i = 0;
        while (1) {
            my $re = $regexps->[$i++];
            $sub = $regexps->[$i++];
            last unless $sub;
            last if @matches = $request =~ /$re/;
        }
    }

    open($s->{stdout}, '>&STDOUT') or die "Can't dup STDOUT: $!";

    $@ = "No such request: $path";
    if (defined $sub) {
        local *STDOUT;
        $s->{output} = '';
        open(STDOUT, '>', \$s->{output});
        eval { &$sub($s, $s->{query}, \@p, \@matches); };
    }
    $| = 1;
    if ($@ =~ /^reply with a (\d+) please/) { # KLUDGE completely crazy
        $s->send_header('text/plain', $1, 'Error');
        print $s->{output};
    } elsif ($@) {
        $s->error($@);
    } else {
        $s->send_header('text/plain', 200, 'OK');
        print $s->{output};
    }

    local $SIG{ALRM} = sub { exit 0 };
    shutdown(STDOUT, 1);
    alarm 10;

    my $buf;
    while (my $got = sysread(STDIN, $buf, 1024)) {
    }

    exit 0;
}

sub unbuffer_stdout {
    my ($s) = @_;
    open(STDOUT, '>&', $s->{stdout}) if $s->{stdout};
}

sub query {
    my ($s, $name) = @_;
    return $s->{query} unless defined $name;
    return (wantarray) ? @{$s->{query}{$name}} : $s->{query}{$name}[0];
}

sub dump_query {
    my ($s) = @_;

    my $q = $s->query;
    if ($q) {
        for my $k (sort keys %$q) {
            for my $v (@{$q->{$k}}) {
                print "$k=$v\n";
            }
        }
    }
}

sub send_header {
    my ($s, @p) = @_;
    return if $s->{sent_header}++;
    $s->{web}->send_header(@p);
}

sub error {
    my ($s, $msg) = @_;
    $s->send_header('text/plain', 500, 'Error');
    print "$s->{state}{HTTP_Path}: Error: $msg\n";
    exit 1;
}

sub rw {
    my ($s) = @_;
    die "Read-only command attempted to modify settings\n"
        if $s->{readonly};
}

sub suffixes {
    my $s = shift;
    for my $suf (@_) {
        return if $suf eq $s->{suffix};
    }
    my $ok = join(', ', map { "'.$_'" } @_);
    die "'.$s->{suffix}' not allowed - must be one of [$ok]";
}

sub sys {
    my ($s, @cmd) = @_;
    my $cmd = join(' ', @cmd);
    if (1 || defined wantarray) { # HIDEOUS KLUDGE
        my $result = `$cmd 2>&1`; # HIDEOUS KLUDGE
        die("Command '$cmd' failed") unless defined $result;
        return $result;
    } else {
        if (system(@cmd)) {
            die("Command '$cmd' failed");
        }
    }
}

sub file_content {
    my ($s, $filename) = @_;
    local $/ = undef;
    open(my $fh, '<', $filename) or die "Can't open $filename: $!";
    my $content = <$fh>;
    close($fh);
    return $content;
}

sub file_find {
    my ($s, $dir, $pattern, $r) = @_;
    $r = [] unless defined $r;
    opendir(my $dh, $dir) or die "Can't open $dir: $!";
    my @e = grep(!/^\.\.?$/, readdir($dh));
    closedir($dh);
    $dir .= '/' unless $dir eq '/';
    map { $_ = "$dir$_" } @e;
    push @$r, @e;
    for (@e) {
        $s->file_find($_, $pattern, $r) if -d && !-l;
    }
    @$r = grep(/$pattern/, @$r) if defined $pattern;
    return $r;
}

sub file_stat {
    my ($s, $path) = @_;

    my @st = stat($path);
    my $r = {
        name=>$path,
        mode=>sprintf("0%o", $st[2]),
        uid=>$st[4],
        gid=>$st[5],
        size=>$st[7],
        #atime=>$st[8],
        modified=>$s->iso_time($st[9]),
        modified_local=>$s->iso_time_local($st[9]),
        #ctime=>$st[10],
    };
    return $r;
}

sub iso_time {
    my ($s, $time, $local) = @_;

    $time = time() unless defined $time;

    my @t = ($local) ? localtime($time) : gmtime($time);
    my $iso = sprintf "%.04d-%.02d-%.02dT%.02d:%.02d:%.02d",
        $t[5] + 1900, $t[4] + 1, $t[3], $t[2], $t[1], $t[0];
    return $iso;
}

sub iso_time_local {
    return iso_time($_[0], $_[1], 1);
}

sub file_newest {
    my ($s, $dir) = @_;

    opendir(my $dh, $dir) or die "Can't open $dir: $!";
    my @e = map { $s->file_stat("$dir/$_") } grep(!/^\.\.?$/, readdir($dh));
    closedir($dh);

    @e = sort { $b->{modified} cmp $a->{modified} } @e;
    return $e[0];
}

sub do_perl {
    my ($s, $pl) = @_;
    return $s->{web}->do_perl($s->{state}{HTTP_Method},
        "$s->{state}{HTTP_Root}$pl", $s->{state}{HTTP_Path});
}

sub emit_dispatch {
    my ($s, $o, $header, $type) = @_;
    if ($type eq 'json') {
        $s->emit_json($o);
        print "\n";
    } elsif ($type eq 'tce') {
        $s->emit_tce($o, $header);
    } elsif ($type eq 'perl') {
        $s->emit_perl($o);
        print "\n";
    }
}

sub emit_json {
    my ($s, $o) = @_;
    my $ref = ref($o);
    if (!$ref) {
        $o =~ s/(['\\])/\\$1/sg;
        $o =~ s/([\x00-\x19])/sprintf("\\u%04X", ord($1))/seg;
        print "'$o'";
    } elsif ($ref eq 'ARRAY') {
        print "[";
        my $comma = 0;
        for my $e (@$o) {
            print($comma++ ? ",\n" : "\n");
            $s->emit_json($e);
        }
        print "\n]";
    } elsif ($ref eq 'HASH') {
        print "{\n";
        my $comma = 0;
        for my $k (sort keys %$o) {
            print($comma++ ? ",\n" : "");
            $s->emit_json($k);
            print ": ";
            $s->emit_json($o->{$k});
        }
        print "\n}";
    }
}

sub emit_perl {
    my ($s, $o) = @_;
    my $ref = ref($o);
    if (!$ref) {
        $o =~ s/(['\\])/\\$1/sg;
        print "'$o'";
    } elsif ($ref eq 'ARRAY') {
        print "[";
        my $comma = 0;
        for my $e (@$o) {
            print($comma++ ? ",\n" : "\n");
            $s->emit_perl($e);
        }
        print "\n]";
    } elsif ($ref eq 'HASH') {
        print "{\n";
        my $comma = 0;
        for my $k (sort keys %$o) {
            print($comma++ ? ",\n" : "");
            $s->emit_perl($k);
            print "=>";
            $s->emit_perl($o->{$k});
        }
        print "\n}";
    }
}

sub emit_tce {
    my ($s, $o, $header) = @_;
    if (ref($o) eq 'HASH') {
        $s->emit_tce_hash($o, $header);
    } elsif (ref($o) eq 'ARRAY') {
        my $idx = 1;
        for my $h (@$o) {
            $s->emit_tce_hash($h, "${header}_$idx");
            ++$idx;
        }
    }
}

sub emit_tce_hash {
    my ($s, $o, $header) = @_;

    return if ref($o) ne 'HASH';
    $header = "Object" unless defined $header;
    print "[$header]\n";
    for my $k (sort keys %$o) {
        $s->emit_tce_element($k, $o->{$k});
    }
    print "\n";
}

sub emit_tce_element {
    my ($s, $tag, $o) = @_;
    my $ref = ref($o);
    if (!$ref) {
        $o =~ s/([\x0D\x0A])/sprintf("%%%02X", ord($1))/eg; # KLUDGE WRONG BAD
        print "$tag=$o\n";
    } elsif ($ref eq 'ARRAY') {
        my $idx = 1;
        for my $e (@$o) {
            $s->emit_tce_element($tag . '_' . $idx++, $e);
        }
    } elsif ($ref eq 'HASH') {
        for my $k (sort keys %$o) {
            $s->emit_tce_element("$tag.$k", $o->{$k});
        }
    }
}

sub sie_file_stat {
    my ($s, $id) = @_;

    if ($id eq '_DEFAULT_') {
        my $list = $s->sie_file_list; # Sorted, newest first
        die "No such file" unless defined $list->[0];
        return $list->[0];
    }

    my $sie = "$SIE_Dir/$id";
    return undef unless -d $sie;
    my $tr = $s->file_stat($sie);
    $tr->{directory} = $sie;
    $tr->{id} = $id;
    $tr->{name} = $s->file_content("$sie/setupname");
    my $si_files = $s->file_find($sie, '\.si$');
    @$si_files > 0 or die "no si files in $tr->{directory}";
    $tr->{size} = 0;
    for my $f (@$si_files) {
        my $si = $s->file_stat($f);
        $tr->{size} += $si->{size};
        $tr->{modified} = $si->{modified}
            if ($tr->{modified} cmp $si->{modified}) < 0;
    }
    for my $limit ([1024*1024*1024, 'GiB'], [1024*1024, 'MiB'],
        [1024, 'KiB']) {
        if ($tr->{size} >= $limit->[0]) {
            $tr->{size_summary} = sprintf("%4.2f %s",
                $tr->{size} / $limit->[0], $limit->[1]);
            last;
        }
    }
    $tr->{si_files} = $si_files;
    my $urlname = $s->{web}->escape_url($tr->{name});
    $tr->{get_url} = "/-/test/$id/data/get/$urlname.sie"; # KLUDGE /sie/ ?
    $tr->{delete_url} = "/~/test/$id/data/delete.txt"; # KLUDGE /sie/ ?
    $tr->{quarantined} = ($id =~ /^[0-9A-Z]{8}$/) ? 0 : 1;
    delete @{$tr}{qw(mode uid gid)}; # KLUDGE lose irrelevant stuff
    return $tr;
}

sub sie_file_list {
    my ($s) = @_;

    opendir(my $dh, $SIE_Dir) or return [];
    my @e = grep(!/^\.\.?$/, readdir($dh));
    closedir($dh);

    my $r = [];

    for my $id (@e) {
        next if $id eq '_DEFAULT_'; # Paranoid infinite-recursion protection
        my $tr = eval { $s->sie_file_stat($id) };
        push(@$r, $tr) if $tr;
    }
    @$r = sort { $b->{modified} cmp $a->{modified} } @$r;
    return $r;
}

sub sie_contents {
    my ($s, $sie_id) = @_;

    my $file = $sie_id;
    unless (-e $file) {
        my $sie_stat = $s->sie_file_stat($sie_id);
        $file = $sie_stat->{si_files}[0];
    }

    require SoMat::SIE;

    my $sie_ctx = SoMat::SIE->new;
    my $sie = $sie_ctx->file_open($file)
        or die "Can't open $file: $!";

    my $out = {ch_id=>{}, ch_prefix=>[], datamode=>[],
        file_tag=>[], run=>[]};

    my (%runs, %datamodes, %ch_prefixes);

    my $sie_channels = $sie->get_channels;
    while (my $ch = $sie_channels->get_next) {
        my $channel_name = $ch->get_name;
        $out->{ch_id}{$channel_name} = $ch->get_id;
        if ($channel_name =~ /^([^\.]*)\.RN_(\d+)$/) {
            $runs{$2}++;
            my $chid = $1;
            my $dm = $chid;
            if ($chid =~ /^([^\@]*)\@/) {
                $dm = $1;
            }
            push @{$out->{datamode}}, $dm unless $datamodes{$dm}++;
            push @{$out->{ch_prefix}}, $chid unless $ch_prefixes{$chid}++;
        }
    }
    $out->{run} = [ sort { $a <=> $b } keys %runs ];

    my $tags = $sie->get_tags;
    while (my $tag = $tags->get_next) {
        push @{$out->{file_tag}}, $tag->name;
    }

    return $out;
}

sub param_comma_list {
    my ($s, $name) = @_;
    my $prm = $s->{state}{HTTP_Post_Hash}{$name};
    $prm ||= $s->{state}{HTTP_Query_Hash}{$name};
    #my $prm = $s->{web}->param($name, 'ALL') || [];
    my @p = map { split(/\s*\,\s*/, $_) } @$prm;
    return \@p;
}

sub send_sie_header {
    my ($s, $filename, $size) = @_;
    my $heads = qq(Content-Disposition: attachment; filename="$filename"\r\n);
    $heads .= qq(Content-Length: $size\r\n) if defined $size;
    # KLUDGE application/sie because Microsoft does a totally
    # stupid attempt at filetype autodetection with
    # application/octet-stream!
    $s->send_header('application/sie', 200, 'OK', $heads);
}

sub sie_extract {
    my ($s, $sie_id, $q) = @_;
    my $sie_stat = $s->sie_file_stat($sie_id);
    my $file = $sie_stat->{si_files}[0];
    my $output_name = "$sie_stat->{name}_extracted.sie";

    $s->unbuffer_stdout;
    require SoMat::SIE;

    my $sie_ctx = SoMat::SIE->new;

    my $runs = $s->param_comma_list('run');
    my $datamodes = $s->param_comma_list('datamode');
    my $channels = $s->param_comma_list('ch_prefix');
    my $chlist = $s->param_comma_list('ch');
    my $globals = $s->param_comma_list('file_tag');
    my $all_globals = $s->param_comma_list('all_file_tags')->[0];

    my $do_sort = 0;
    for my $sorts ($runs, $datamodes, $channels, $chlist) {
        ++$do_sort if @$sorts;
    }
    $do_sort = 0 unless $do_sort > 1;

    my $restrictors = (@$runs || @$datamodes || @$channels);

    $runs = ['ALL'] unless @$runs;
    $channels = ['ALL'] unless @$channels;
    $datamodes = ['ALL'] unless @$datamodes;

    my @sies;
    my @chlist;
    my $sie_map = {};

    for my $fname ($file) {
        my $sie = $sie_ctx->file_open($fname)
            or die "Can't open $fname: $!";

        push @chlist, map { [scalar @sies, $_] } @$chlist;

        my $sie_ch = [];
        my $sie_channels = $sie->get_channels;
        while (my $ch = $sie_channels->get_next) {
            push @$sie_ch, $ch->get_name;
            $sie_map->{scalar @sies}{$ch->get_name} = $ch->get_id;
        }

        my $last_run = 0;
        for my $ch (@$sie_ch) {
            if ($ch =~ /\.RN_(\d+)$/) {
                $last_run = $1 if $last_run < $1;
            }
        }

        my @run_pat = ('.*');
        @run_pat = map {
            my $run = $_;
            $run = $last_run + 1 + $run if $run < 0;
            ".RN_$run\$"
        } @$runs unless $runs->[0] eq 'ALL';

        my @dm_pat = ('.*');
        @dm_pat = map { "^$_\[\\\@\\\.\]" } @$datamodes
            unless $datamodes->[0] eq 'ALL';

        my @chan_pat = ('.*');
        @chan_pat = map { "\\b$_\\." } @$channels
            unless $channels->[0] eq 'ALL';

        #print " run_pat: ", join("|", @run_pat), "\n";
        #print "  dm_pat: ", join("|", @dm_pat), "\n";
        #print "chan_pat: ", join("|", @chan_pat), "\n";

        if ($restrictors) {
            for my $ch (@$sie_ch) {
                my $ok = 0;
                for my $p (@run_pat) {
                    $ok |= 1 if $ch =~ /$p$/;
                }
                for my $p (@chan_pat) {
                    $ok |= 2 if $ch =~ /$p/;
                }
                for my $p (@dm_pat) {
                    $ok |= 4 if $ch =~ /^$p/;
                }
                next unless $ok == 7;
                push @chlist, [scalar @sies, $ch];
            }
        }
        push @sies, $sie;
    }
    @chlist = sort {
        my $cmp = $a->[0] <=> $b->[0]; 
        return $cmp if $cmp;
        my $a1 = $a->[1];
        my $b1 = $b->[1];
        my ($ar) = $a1 =~ /\.RN_(\d+)$/;
        my ($br) = $b1 =~ /\.RN_(\d+)$/;
        $cmp = $ar <=> $br;
        return $cmp if $cmp;
        return $a1 cmp $b1;
    } @chlist if $do_sort;

    my $fh = \*STDOUT;
    my $omit_content_length = $q->{omit_content_length};
    $q->{omit_content_length} = 0 unless defined $q->{omit_content_length};
    my $of = $q->{output};
    $of = '' unless defined $of;
    my $is_file = ($of ne '');
        
    my $err = undef;
    if ($of =~ /^.+\.gz$/ || lc($q->{compress}) eq 'gzip') {
        local $SIG{PIPE} = sub { $err = "Write failed" };
        $of =~ s/'/\\'/g;
        my $pipe_cmd = "|exec /bin/gzip -9" . ($is_file ? " > '$of'" : "");
        $fh = SoMat::File->new($pipe_cmd);
        $omit_content_length = 1;
    } elsif ($is_file) {
        $fh = SoMat::File->new(">$of");
    }
    die "Can't open $of: $!" unless defined $fh;

    my $sie_bytes = 0;
    my $out_buf = ($is_file) ? undef : '';

    my $writer = $sie_ctx->new_writer(
        sub {
            my ($data) = @_;
            if (defined $out_buf) {
                # Buffer the data until $sie_bytes has been determined
                # unless $omit_content_length is set. Then emit the HTTP
                # header and accumulated data, then disable buffering.
                if ($sie_bytes || $omit_content_length) {
                    $s->send_sie_header($output_name,
                        $omit_content_length ? undef : $sie_bytes);
                    print $fh $out_buf, $data;
                    $out_buf = undef;
                } else {
                    $out_buf .= $data;
                    # Try to avoid exhausting memory with a sensible limit
                    $omit_content_length = 1 if length($out_buf) > 2000000;
                }
            } else {
                print $fh $data;
            }
            return length $data;
        }
    );
    $writer->xml_header;
    my $sifter = $writer->new_sifter;

    push @$globals, 'core:version', 'somat:version';

    if ($all_globals || @$globals) {
        my %map;
        $map{$_} = 1 for @$globals;
        my $sie = $sies[0]; # KLUDGE!!!
        my $tags = $sie->get_tags;
        while (my $tag = $tags->get_next) {
            if ($all_globals || $map{$tag->name}) {
                $sifter->add($tag);
            }
        }
    }

    my $tagged_tests = {};

    for my $ch (@chlist) {
        my $sie = $sies[$ch->[0]];
        my $siech = $sie->get_channel($sie_map->{$ch->[0]}{$ch->[1]});
        my $test = $siech->get_containing_test;
        unless ($tagged_tests->{$test->get_id}++) {
            my $tags = $test->get_tags;
            while (my $tag = $tags->get_next) {
                $sifter->add($tag);
            }
        }
        $sifter->add($siech);
    }

    $sie_bytes = $sifter->total_size;

    # KLUDGE until $sifter->finish works properly
    undef $sifter;
    undef $writer;

    if ($of ne '') {
        close($fh) or die "Error closing $of: $!";
    }
    $@ = undef; # KLUDGE???
}

sub get_checksum {
    my ($file, $type) = @_;
    my $sum;
    open($sum, "${type}sum " . quotemeta($file) . "|");
    my $csum_line = <$sum>;
    my $csum = $1 if ($csum_line =~ /^([0-9a-f]+)/);
    die "${type}sum failed on file $file" unless $csum;
    return $csum;
}

1;
