use strict;

package SoMat::eDAQ::WebServer;

use Socket;
use Symbol ();

my $MAX_CONTENT = 4000000;
my $WebServer_State;

my $CONTENT_TYPES = {
    html => 'text/html',
    js => 'text/javascript',
    css => 'text/css',
    gif => 'image/gif',
    jpg => 'image/jpeg',
    png => 'image/png',
};

sub new {
    my $class = shift;
    my $s = bless { @_ }, ref($class) || $class;
    $s->make_state;
    $s->{state}{pastbuf} = '' unless defined $s->{state}{pastbuf};
#print STDERR "WebServer::new() state=$s->{state}\n";
    unless ($s->{no_read_post}) {
        if ($s->is_multipart) {
#print STDERR "    calling read_multipart()\n";
            $s->read_multipart;
        } else {
#print STDERR "    calling read_post()\n";
            $s->{state}->{HTTP_Content}
                = $s->read_post($s->{state}->{HTTP_Headers});
#print STDERR "    content=($s->{state}{HTTP_Content})\n";
            $s->{state}->{HTTP_Post_Hash}
                = $s->decode_query($s->{state}->{HTTP_Content});
        }
    }
    return $s;
}

sub preload {
    my ($s, $dir) = @_;
    $dir = '/usr/local/custom' unless defined $dir;
    my $dh = Symbol::gensym();
    if (opendir($dh, $dir)) {
        for my $de (readdir($dh)) {
            next if $de eq '.' || $de eq '..';
            my $f = "$dir/$de/.web_preload";
            do $f if -f $f;
        }
        closedir $dh;
    }
    my $f = '/etc/web_preload.pl';
    do $f if -f $f;
}

sub get {
    my ($s, $name) = @_;
    return $s->{state} unless defined $name;
    return $s->{state}->{$name};
}

sub set {
    my ($s, $name, $val) = @_;
    if (ref($name) eq 'HASH') {
        $s->{state} = $name;
        return;
    }
    $s->{state}->{$name} = $val;
}

sub make_state {
    my ($s) = @_;
    $WebServer_State ||= {};
    return $s->{state} = $WebServer_State;
}

# SoMat::eDAQ::WebServer->serve('/usr/local/web/public');
sub serve {
    my ($class, $root, $buf) = @_;
    $root = '/usr/local/web/public' unless defined $root;

open(STDERR, '>>', '/dev/null');
#open(STDERR, '>>', '/hd/webinet.stderr');
select((select(STDERR), $|=1)[0]);
#print STDERR "\n";
#printf STDERR "WebServer::serve(root=$root), length(buf)=%d\n", length($buf);

    # KLUDGE necessary since edaqserver changed to use sigaction(),
    # otherwise `ls` will return undef with system call interrupted...
    $SIG{CHLD} = 'DEFAULT';

    #$s = SoMat::eDAQ::WebServer->new(); # KLUDGE use global?
    my $s = bless { }, ref($class) || $class;
    my $state = $s->make_state;

    $s->{state}{pastbuf} = ''; # KLUDGE do in make_state
    $s->{state}{inbuf} = $buf;

#printf STDERR "WebServer::serve before read_request_line, length(inbuf)=%d\n", length($s->{state}{inbuf});
    my ($method, $path) = $s->read_request_line; # KLUDGE?
#printf STDERR "WebServer::serve after read_request_line, length(inbuf)=%d\n", length($s->{state}{inbuf});

    # KLUDGE this is gawdawful, sanify proxying somehow
    if ($path =~ m{^/\~/tce/(version|query|init_node|dvm|cal|rezero|prerun_rezero|abort|aux_files)}) {
        $s->tce_proxy($method, $path);
    }

    my $headers = $s->read_headers;

    $path = $s->{state}{HTTP_Path}; # KLUDGE
    $state->{HTTP_Root} = $root;

#printf STDERR "    X3 length(inbuf)=%d\n", length($s->{state}{inbuf});

#printf STDERR "    X4 length(inbuf)=%d\n", length($s->{state}{inbuf});
    if ($path =~ /^\/[\-\~]\//) {
        require SoMat::eDAQ::Web::Commands;
        SoMat::eDAQ::Web::Commands->run($s);
    }
#printf STDERR "    X5 length(inbuf)=%d\n", length($s->{state}{inbuf});

    if ($path =~ /^(.*?\.cgi)(.*)$/ && -x "$root/$1") {
        $ENV{REQUEST_METHOD} = $method;
        $ENV{SCRIPT_NAME} = $1;
        $ENV{PATH_INFO} = $2;
        $ENV{PATH} = '/usr/local/bin:/usr/bin/:/bin:/usr/sbin:/sbin';
        $ENV{QUERY_STRING} = $state->{HTTP_Query};
        $ENV{CONTENT_LENGTH} = $headers->{'content-length'};
        $ENV{CONTENT_TYPE} = $headers->{'content-type'};
        exec "$root/$1";
    }
#printf STDERR "    X6 length(inbuf)=%d\n", length($s->{state}{inbuf});


    $path =~ s/(\/\.\.?)$/$1\//;
    # Kill this-directory refs
    $path =~ s/\/\.\//\//g;
    # Kill multi-slashes
    $path =~ s/\/+/\//g;
    # Kill up-directory refs
    1 while $path =~ s/\/(?!\.\.\/)([^\/]*\/)\.\.//g;
    # Kill other characters we don't like in file/directory names
    #$path =~ s/[^\w\-\/\.\$]//g;
    # Kill trailing slashes
    $path =~ s/\/+$//;

    my $p = "$root$path";
    return $s->send_file($p) if $method eq 'GET' && -f $p && $p !~ /\.pl$/;

    my @p = split(/\//, $path);

#printf STDERR "    X7 length(inbuf)=%d\n", length($s->{state}{inbuf});

    push @p, ''; # We want to start at _directory_ $dir/$file...
    while (@p) {
        my $file = pop(@p);
        my $dir = join('/', @p);
        next unless -d "$root/$dir";
        $file =~ s/\.\w+$//;
        my $u = "$root$dir/$file.pl";
        return 0 if $file ne 'index' && -f $u
            && $s->do_perl($method, $u, "$dir/$file");
        $u = "$root$dir/index.pl";
        return 0 if -f $u && $s->do_perl($method, $u, $dir);
    }

    $p =~ s/\/+$//;

    if (-d $p) {
        my $t;
        if (-f ($t = "$p/index.pl")) {
            return 0 if $s->do_perl($method, $t, $p);
        } elsif (-f ($t = "$p/index.html")) {
            $p = $t;
        }
    }
    if ($method eq 'GET' || $method eq 'POST') { # KLUDGE Accept more?
        return $s->send_file($p) if -f $p;
        return $s->http_error(404, "Not Found: $path");
    }
    return $s->http_error(405, "Method $method Not Allowed");
}

sub do_perl {
    my ($s, $method, $p, $script_path) = @_;
    if ($method eq 'GET' || $method eq 'POST') { # KLUDGE Accept more?
        my $dir = $p;
        $dir =~ s/\/[^\/]*$//;
        $s->{state}->{HTTP_ScriptPath} = $script_path;
        my $r = eval
            "package main; chdir '$dir'; my \$r = do '$p';"
            . " return undef if defined \$r; return join(': ', \$@, \$!);";
        $s->http_error(500, $r) if defined $r && $r !~ /^PAGE_ABORT/;
        return 1;
    }
    return 0;
}

sub rawread {
    my ($s, undef, $length, $offset) = @_;

    my $fh = \*STDIN;
    $offset ||= 0;

#printf STDERR "rawread, inbuf length=%d\n", length($s->{state}{inbuf});
    if (my $hblen = length($s->{state}{inbuf})) {
        $length = $hblen if $length > $hblen;
        substr($_[1], $offset, length($_[1]) - $offset,
            substr($s->{state}{inbuf}, 0, $length, ''));
        return $length;
    }
#printf STDERR "rawread, calling sysread, length=$length, offset=$offset\n";
    my $got = sysread($fh, $_[1], $length, $offset);
    return $got;

#printf STDERR "rawread(fh=$fh, length=$length, offset=$offset), fd=%d\n", fileno($fh);
    while (1) {
        my $got = sysread($fh, $_[1], $length, $offset);
        # KLUDGE - need $! check
        return $got if defined $got;
#printf STDERR "rawread(), sysread returned UNDEF: $! = %d\n", $!;
        next if $! == 11; # KLUDGE Errno::EAGAIN() - nonblocking I/O
        next if $! == 4; # KLUDGE Errno::EINTR()
        return undef;
    }
}

sub get_line {
    my ($s) = @_;
#printf STDERR "get_line IN, inbuf length=%d\n", length($s->{state}{inbuf});
    while (length($s->{state}{inbuf}) < 16384) {
#printf STDERR "get_line LOOP, inbuf length=%d\n", length($s->{state}{inbuf});
        if ($s->{state}{inbuf} =~ s/^((.*?)\x0D\x0A)//) {
            $s->{state}{pastbuf} .= $1;
#printf STDERR "get_line RETURNING, inbuf length=%d\n", length($s->{state}{inbuf});
            return $2;
        }
#printf STDERR "get_line calling sysread(), inbuf length=%d\n", length($s->{state}{inbuf});
        my $got = sysread(\*STDIN, $s->{state}{inbuf}, 1024, length($s->{state}{inbuf}));
#printf STDERR "get_line after sysread, got=$got, inbuf length=%d\n", length($s->{state}{inbuf});
        if ($got <= 0) {
            return undef;
        }
    }
    # KLUDGE return 'overlength header line' error?
    return undef;
}

sub read_request_line {
    my ($s) = @_;
    my $line = $s->get_line;
#print STDERR "read_request_line A line=($line)\n";
    return undef unless defined $line;
    # KLUDGE should do this better;
#print STDERR "read_request_line B line=($line)\n";
    my ($method, $path) = ($line =~ /^(\w+)\s+([^\s]*)/);
    $method = uc($method); # KLUDGE?
    $path =~ s|^http://([^/])*/|/|i;
#print STDERR "read_request_line C method=($method), path=($path)\n";

    my $state = $s->{state};
    $state->{HTTP_Method} = $method;
    $state->{HTTP_PathAndQuery} = $path;

    my $iaddr;
    if (defined (my $sname = getsockname(\*STDIN))) { # KLUDGE
        ($state->{HTTP_Port}, $iaddr) = sockaddr_in($sname);
        $state->{HTTP_Host} = $state->{HTTP_Address} = inet_ntoa($iaddr);
    }

    # From page 50 of http://www.ietf.org/rfc/rfc3986.txt
    #my $uri_re = qr{^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?};
    ##                12        2 1 3  4       43 5      56  7     76 8 9  98
    #my $scheme    = $2
    #my $authority = $4
    #my $path      = $5
    #my $query     = $7
    #my $fragment  = $9

    if ($path =~ /^([^\?]*)(\?(.*))?$/) {
        $state->{HTTP_Path} = $1;
        $state->{HTTP_Query} = $3;
#print STDERR "read_request_line HTTP_Path=$1, HTTP_Query=$3\n";
        $state->{HTTP_Query_Hash} = $s->decode_query($3);
    }

    return ($method, $path);
}

sub read_headers {
    my ($s) = @_;
    my %headers;
    my $hdr = 'UNDEF';
    while (1) {
        my $line = $s->get_line;
        # KLUDGE do something if undef?
        die "Bad input header $hdr" unless defined $line;
        last unless length($line);
        if ($line =~ /^([\w\-]+)\:\s*(.*)$/) {
            $hdr = lc($1);
            $headers{$hdr} = $2;
        } elsif ($line =~ /^\s/ && defined $hdr) {
            $headers{$hdr} .= $line;
        } else {
            die "Bad input header $hdr";
        }
    }
#print STDERR "read_headers, method=$method, path=$path, headers:\n";
#for my $k (sort keys %headers) { print STDERR "  $k => $headers{$k}\n" }
    if (exists $headers{host}) {
        ($s->{state}{HTTP_Host} = $headers{host}) =~ s/:\d*$//; # Kill off port 
    }
    $s->{state}{HTTP_Headers} = \%headers;
    return \%headers;
}

sub send_header {
    my ($s, $ctype, $code, $msg, $extras) = @_;
    $ctype = "text/plain" unless defined $ctype;
    $code = 200 unless defined $code;
    $msg = "OK" unless defined $msg;
    $extras = "" unless defined $extras;
    unless ($extras =~ /^Date:/mi) {
        my $date = scalar(localtime);
        $extras .= 
            "Date: $date\r\n" .
            "Expires: $date\r\n" .
            "Pragma: no-cache\r\n" .
            "Cache-Control: no-cache\r\n";
    }
    syswrite STDOUT,
        "HTTP/1.0 $code $msg\r\n" .
        "Content-type: $ctype\r\n" .
        "Server: edaqserver\r\n" .
        "Connection: close\r\n" .
        $extras .
        "\r\n";
}

my $HTTP_Status_Codes = {
    100=>'Continue',
    101=>'Switching Protocols',
    200=>'OK',
    201=>'Created',
    202=>'Accepted',
    203=>'Non-Authoritative Information',
    204=>'No Content',
    205=>'Reset Content',
    206=>'Partial Content',
    300=>'Multiple Choices',
    301=>'Moved Permanently',
    302=>'Found',
    303=>'See Other',
    304=>'Not Modified',
    305=>'Use Proxy',
    307=>'Temporary Redirect',
    400=>'Bad Request',
    401=>'Unauthorized',
    402=>'Payment Required',
    403=>'Forbidden',
    404=>'Not Found',
    405=>'Method Not Allowed',
    406=>'Not Acceptable',
    413=>'Request Entity Too Large',
    500=>'Server Error',
};

sub http_error {
    my ($s, $code, $details, $msg) = @_; # $msg === Reason-Phrase
    if (@_ == 4 && $_[2] =~ /^\d+$/) { # KLUDGE Swap for legacy call order
        ($s, $msg, $code, $details) = @_;
    }
    $msg ||= $HTTP_Status_Codes->{$code};
    $details ||= $msg;
    $s->send_header("text/html", $code, $msg);
    syswrite STDOUT,
        "<html>\n" .
        "<head>\n" .
        "<title>$code $msg</title>\n" .
        "</head>\n" .
        "<body bgcolor=\"#FFFFFF\">\n" .
        "HTTP Error $code: $msg" .
        (defined($details) ? "<p>\n$details\n" : "") .
        "</body>\n" .
        "</html>\n";
    return undef;
}

sub send_file {
    my ($s, $path, $ctype, $heads) = @_;
    $heads = "" unless defined $heads;
    unless (defined $ctype) {
        $ctype = "text/plain";
        if ($path =~ /\.(\w+)$/) {
            my $suffix = lc($1);
            $ctype = $CONTENT_TYPES->{$suffix} if $CONTENT_TYPES->{$suffix};
        }
    }
    my $fh = Symbol::gensym;
    sysopen($fh, $path, 0)
        or return $s->http_error(404, "Can't open $path: $!");
    my @stat = stat($fh);
    my $clen = $stat[7];
    $clen = undef if -b $fh || -c $fh; # No Content-Length for device files
    my $mtime = localtime($stat[9]);
    $heads .= "Date: $mtime\r\n" unless $heads =~ /^Date:/mi;
    $heads .= "Content-Length: $clen\r\n"
        if defined $clen && $heads !~ /^Content-Length:/mi;
    $s->send_header($ctype, 200, "OK", $heads);
    my $buf;
    $| = 1;
    print STDOUT ''; # Flush before doing system I/O
    syswrite(STDOUT, $buf) while sysread($fh, $buf, 10000) > 0;
    close($fh);
}

sub read_post {
    my ($s, $h) = @_;
    my $data = '';
    my $len = $h->{'content-length'};

#print STDERR "read_post, content-length=$len\n";
    return undef unless $len;
    return $s->http_error(413) if $len > $MAX_CONTENT;
    my $off = 0;
    while ($len > 0) {
#print STDERR "read_post, calling rawread(), len=$len\n";
        my $rlen = $s->rawread($data, $len, length($data));
        last unless $rlen > 0;
        $len -= $rlen;
    }
#printf STDERR "read_post, length=%d\n", length($data);
    return undef unless length($data) == $h->{'content-length'};
    return $data;
}

sub close_connection {
    my ($s) = @_;

    # KLUDGE make this a method...    
    eval {
        local $SIG{ALRM} = sub { die };
        shutdown(STDOUT, 1); # KLUDGE STDOUT
#print STDERR "after shutdown\n";
        alarm 30;
        my $buf;
        while (my $got = sysread(STDIN, $buf, 1024)) { # KLUDGE STDIN
#print STDERR "post-shutdown got $got extra bytes\n";
        }
        alarm 0;
    };
}

sub decode_query {
    my ($s, $query, $hashref) = @_;
    $hashref = {} unless defined $hashref;
    map {
        my ($key, $val) = split('=', $_, 2);
#print STDERR "decode_query, key=$key, val=$val\n";
        push(@{$hashref->{$s->unescape_url($key)}}, $s->unescape_url($val));
    } split(/[\&\;]/, $query);
    return $hashref;
}

sub _param {
    my ($s, $params, $name, $pos) = @_;
    return $params unless defined $name;
    $pos = 0 unless defined $pos;
    return undef unless exists $params->{$name};
    return $params->{$name} if $pos eq 'ALL';
    return $params->{$name}->[$pos];
}

sub param { 
    my $p = &query_param;
#print STDERR "param(name=$_[1])\n";
#print STDERR "param(name=$_[1]) val=$p\n" if defined $p;
    return $p if defined $p;
    my ($s, $name, $pos) = @_;
    return $s->_param($s->{state}->{HTTP_Post_Hash}, $name, $pos);
}

# Returns stuff from QUERY_STRING
sub query_param { 
    my ($s, $name, $pos) = @_;
    return $s->_param($s->{state}->{HTTP_Query_Hash}, $name, $pos);
}

sub post_param { 
    my ($s, $name, $pos) = @_;
    return $s->_param($s->{state}->{HTTP_Post_Hash}, $name, $pos);
}

sub full_url {
    my ($s, $url) = @_;
    if ($url !~ /^http/) {
        if ($url =~ /^([\:\/])/) {
            my $port = "";
            $port = ':' . $s->{state}->{HTTP_Port}
                if $1 ne ':' && $s->{state}->{HTTP_Port} != 80; 
            my $host = $s->{state}->{HTTP_Address};
            $host = $s->{state}->{HTTP_Host}
                if exists $s->{state}->{HTTP_Host};
            $url = "http://$host$port$url";
        } else {
            # KLUDGE do something here for relative paths from $s
        }
    }
    return $url;
}

sub script_path {
    my ($s) = @_;
    return $s->{state}->{HTTP_ScriptPath};
}

sub redirect {
    my ($s, $url) = @_;
    $url = $s->full_url($url);
    syswrite STDOUT,
        "HTTP/1.0 302 Redirect\r\n" .
        "Location: $url\r\n\r\n";
    return 0;
}

sub escape_url {
    my ($s, $str) = @_;
    return undef unless defined $str;
    $str =~ s/([^A-Za-z0-9\-_.!~*'()])/sprintf("%%%02X", ord($1))/eg;
    return $str;
}

sub unescape_url {
    my ($s, $str) = @_;
    $str =~ tr/+/ /;
    $str =~ s/%([0-9a-fA-F]{2})/pack("c", hex($1))/ge;
    return $str
}

# Stolen/Modified from CGI.pm
sub escape_html {
    my ($s, $str) = @_;
    return "" unless defined $str;

    $str =~ s/&/&amp;/g;
    $str =~ s/\"/&quot;/g;
    $str =~ s/>/&gt;/g;
    $str =~ s/</&lt;/g;
    return $str;
}

# Stolen/Modified from CGI.pm
sub unescape_html {
    my ($s, $str) = @_;
    return undef unless defined $str;
    # thanks to Randal Schwartz for the correct solution to this one
    $str=~ s[&(.*?);]{
        local $_ = $1;
        /^amp$/i        ? "&" :
        /^quot$/i       ? '"' :
        /^gt$/i         ? ">" :
        /^lt$/i         ? "<" :
        /^#(\d+)$/      ? chr($1) :
        /^#x([0-9a-f]+)$/i ? chr(hex($1)) :
        $_
        }gex;
    return $str;
}

sub is_multipart {
    my ($s) = @_;

    return undef unless $s->get('HTTP_Method') eq 'POST';
    my $h = $s->get('HTTP_Headers');
    return undef unless $h->{'content-type'} =~ m|^multipart/form-data|;
    return 1;
}

my $MultipartNotice = <<'EOT';
* Multipart read support. Excerpted from CGI.pm 1.30 and heavily modified:
*
* Copyright 1995-1998 Lincoln D. Stein.  All rights reserved.
* It may be used and modified freely, but I do request that this copyright
* notice remain attached to the file.  You may modify this module as you 
* wish, but if you redistribute a modified version, please attach a note
* listing the modifications you have made.
*
* The most recent version of CGI.pm and complete docs are available at:
*   http://stein.cshl.org/WWW/software/CGI/
EOT

sub read_multipart {
    my ($s) = @_;

    return unless $s->is_multipart;

    require SoMat::eDAQ::Web::Multipart;
    require SoMat::File;

    my $CRLF = "\015\012";

    my $h = $s->get('HTTP_Headers');
    my ($boundary) = $h->{'content-type'} =~ /boundary=\"?([^\";,]+)\"?/;
#print STDERR "read_multipart, boundary = $boundary\n";

    my $buffer = SoMat::eDAQ::Web::Multipart->new($s, $boundary,
        $h->{'content-length'});
#print STDERR "read_multipart, buffer = $buffer\n";
    return unless $buffer;

    my $prm = $s->{state}->{HTTP_Post_Hash} = {};

    while (!$buffer->eof) {

#print STDERR "read_multipart, LOOP\n";
        my $header = $buffer->readHeader;
#print STDERR "read_multipart, header = $header\n";
        return $s->http_error(400, "Malformed multipart POST")
            unless defined $header;

        my ($param) = $header->{'content-disposition'} =~ / name="?([^\";]*)"?/;

        # Bug:  Netscape doesn't escape quotation marks in file names!!!
        my ($filename) = $header->{'content-disposition'}
            =~ / filename="?([^\"]*)"?/;
#print STDERR "read_multipart, param = $param, filename = $filename\n";

        # If no filename specified, then just read the data and assign it
        # to our parameter list.
        if ((!defined($filename) || $filename eq '')) {
            my ($value) = $buffer->readBody;
            push(@{$prm->{$param}}, $value);
            next;
        }

        # skip the file if uploads disabled
        #if ($DISABLE_UPLOADS) {
        #    1 while defined $buffer->read;
        #    next;
        #}

        my $tmpfile =
            SoMat::File::Temp->tempfile('tmpfileXXX', $s->{tempfile_dir});
#print STDERR "read_multipart, tmpfile = $tmpfile\n";
        my $fh = SoMat::File->new(">$tmpfile");
        die "read_multipart can't open $tmpfile: $!\n" unless defined $fh;

        local $\ = undef;
        my $data;
        while (defined($data = $buffer->read)) {
#printf STDERR "read_multipart, read %d bytes\n", length($data);
            print $fh $data;
        }
        seek($fh, 0, 0);
        close $fh;

#printf STDERR "read_multipart, DONE with part, name=$param, filename=$filename, file=$tmpfile, info=$header\n";

        my $stuff = {
            name => $param,
            filename => $filename,
            file => $tmpfile,
            info => $header,
        };
        push(@{$prm->{$param}}, $stuff);
    }
#printf STDERR "read_multipart, ALL DONE\n";
    return 1;
}

sub test_status {
    my ($s, %args) = @_;
    my $edaq = $args{edaq};
    $edaq = '127.0.0.1' unless defined $edaq;
    return $s->{test_status}->{$edaq}
        if exists $s->{test_status}->{$edaq} && !$args{force};
    require SoMat::Network::WebRequest;
    my $url = "http://$edaq\:2345/teststatus";
    my $c = {};
    eval {
        my $web = SoMat::Network::WebRequest->new($url);
        while (defined (my $line = $web->getline)) {
            $c->{$1} = $2 if $line =~ /^(\w+)=([^\r\n]*)/;
        }
        if (exists $c->{SetupFileLong}) {
            ($c->{SetupName} = $c->{SetupFileLong}) =~ s/\.tce$//i;
        }
    };
    return ($s->{test_status}->{$edaq} = $c);
}

sub tce_proxy {
    my ($s, $method, $path) = @_;

    $| = 1;
#print "tce_proxy, starting up, method=$method, path=$path...\n";

    my $remote = 'localhost';
    my $port = 7733;
    my $iaddr = inet_aton($remote) or die "no host: $remote";
    my $paddr = sockaddr_in($port, $iaddr);
    my $proto = getprotobyname('tcp');
    my $tce;
    socket($tce, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
    connect($tce, $paddr) or die "connect: $!";
    select((select($tce), $| = 1)[0]);

#print "tce_proxy pastbuf=$s->{state}{pastbuf}\n";
#print "tce_proxy inbuf=$s->{state}{inbuf}\n";

    my $req = "$method $path HTTP/1.0\x0D\x0A$s->{state}{inbuf}";
    print $tce $req;

#print "tce_proxy WROTE(TCE) $req\n\n";
#printf "tce_proxy CLIENT fileno=%d\n", fileno(STDIN);
#printf "tce_proxy TCE fileno=%d\n", fileno($tce);

    # KLUDGE do except?

    my ($r_in, $r_out);
    vec($r_in, fileno(STDIN), 1) = 1;
    vec($r_in, fileno($tce), 1) = 1;

#printf "tce_proxy r_in=%s\n", unpack('H*', $r_in);

    while (1) {
#print "tce_proxy calling select\n";
        my $count = select($r_out = $r_in, undef, undef, undef);
#print "tce_proxy back from select, count=$count\n";
        last if $count < 1;
        my $pbuf;
        my $got;
        if (vec($r_out, fileno(STDIN), 1)) {
#print "tce_proxy CLIENT readable, calling sysread()\n";
            $got = sysread(STDIN, $pbuf, 8192);
#print "tce_proxy CLIENT readable, got=$got\n";
            if ($got > 0) {
                print $tce $pbuf;
            } else {
                vec($r_in, fileno(STDIN), 1) = 0;
            }
        }
        if (vec($r_out, fileno($tce), 1)) {
#print "tce_proxy TCE readable, calling sysread()\n";
            $got = sysread($tce, $pbuf, 8192);
#print "tce_proxy TCE readable, got=$got\n";
            if ($got > 0) {
                print STDOUT $pbuf;
            } else {
                vec($r_in, fileno($tce), 1) = 0;
                last;
            }
        }
    }

    $s->close_connection;

#print "tce_proxy DONE, exiting\n";
    exit 0;
}

1;
