use strict;
use warnings 'all';

# my $s; my $bundle_file; # KLUDGE uncomment for 'perl -c' syntax checks

my $title = 'eDAQ Firmware Updater';

my $status = eval { $s->test_status } || {};
if ($status->{RunStarted} || $status->{RunPreview}) {
    $s->edaq_header($title);
    print "<h3>$title</h3>\n";
    print "<p>eDAQ firmware updates are not permitted while a test run or"
        . " preview is in progress</p>";
    my $f = $s->form;
    print $f->begin;
    print $f->submit(name=>'_Cancel', value=>"OK"); 
    print $f->end;
    return $s->edaq_footer();
}

my $update_dir = '/tmp/updater';
if ($bundle_file !~ /^$update_dir\// || !-d $update_dir) {
    system('/bin/rm', '-rf', $update_dir); # ignore errors
    system('/bin/mkdir', '-p', $update_dir)
        and die "Can't mkdir $update_dir: $!";
    my $nf = "$update_dir/bundle.zip";
    rename($bundle_file, $nf) or die "Can't move $bundle_file: $!";
    $bundle_file = $nf;
    my $cmd = "/usr/bin/unzip -o '$bundle_file' '*/updater/*' -d $update_dir";
    my $out = `$cmd 2>&1`;
}

unshift(@INC, "$update_dir/lib/updater");

require SoMat::DB::Tab;
require SoMat::Network::ProcessMonitor;
require SoMat::eDAQ::Hardware::MultiUpdate;

my $scripts = <<'EOT';
<style type="text/css">

.progress {
    margin: 1em;
}

.progressbarbg {
    background-color: #CCCCCC;
    border-style: solid;
    border-width: .3ex;
    border-color: #FF0000;
    padding: 0px;
    text-align: left;
    width: 30ex;
}

.progressbar {
    background-color: #AA4444;
    border-style: solid;
    border-width: 0px;
    border-color: black;
}

.progress-text {
    display: inline;
}

.progress-text .percent {
    width: 3em;
    display: inline-block;
    text-align: right;
    padding-right: .3em;
}

.progress-text .time {
    padding-right: .3em;
}

.progress-text .name {
    font-weight: bold;
}

.cometsupport {
    width: 1000px;
    height: 400px;
    display: none;
}

.cometsupport iframe {
    width: 1000px;
    height: 400px;
}

tt {
    font-family: monospace;
}

.error {
    color: #cd0a0a;
}

.message {
}

.errorbox {
    border: 1px solid #cd0a0a;
    background: #fef1ec;
    color: #cd0a0a;
    padding: 1em;
    margin-top: 1em;
}

</style>
<script type="text/javascript" src="/js/jquery-1.3.min.js"></script>

<script type="text/javascript">

var edaq_reset_interval = null;
var edaq_reset_wait = 180; // 3 minutes
var edaq_reset_clock = 0;

function format_time(secs) {
    var min = (secs / 60).toString().replace(/\..*$/g, ""); // i.e. floor()
    var sec = (secs % 60).toFixed(0).toString().replace(/^(\d)$/g, "0$1");
    return "[" + min + ":" + sec + "]";
}

var edaq_reset_updater = function() {
    var spec = "#progress #SystemReset ";
    if (++edaq_reset_clock > edaq_reset_wait) {
        clearInterval(edaq_reset_interval);
        $(spec + ".message").text("eDAQ reset wait timed out");
    }
    $.ajax({
        type: "GET",
        url: "/-/test/_DEFAULT_/status.txt",
        dataType: "text",
        timeout: 800,
        success: function(data, textStatus) {
            clearInterval(edaq_reset_interval);
            $(spec + ".percent").text("100%");
            $(spec + ".message").text("Update complete");
            $(spec + ".progressbar").css("width", "100%");
        }
    });
    var pct = 100 * edaq_reset_clock / edaq_reset_wait;
    $(spec + ".time").text(format_time(edaq_reset_clock));
    $(spec + ".percent").text('');
    $(spec + ".message").text("Waiting for eDAQ to reset");
    $(spec + ".progressbar").css("width", pct.toFixed(1) + "%");
}

function comet_update_progress(e) {
    var name = e.name.replace(/[^a-zA-Z0-9]/g, '');
    var spec = "#progress #" + name;
    if ($(spec).size() == 0) {
        $("#progress").append(
            '<div id="' + name + '" class="progress">'
            + '<div id="' + name + '" class="progressbarbg">'
            + '<div class="progressbar" width="0%">&nbsp;</div>'
            + '</div>'
            + '<div class="progress-text">'
            + '<span class="percent">0%</span>'
            + ' <span class="time">[0:00]</span>'
            + ' <span class="name">' + e.name + '</span>'
            + ' &#150; <span class="message"></span>'
            + ' <span class="error"></span>'
            + '</div>'
            + '</div>'
        );
    }
    if (name == "SystemReset" && e.progress == 1) {
        if (edaq_reset_interval == null)
            edaq_reset_interval = setInterval(edaq_reset_updater, 1000);
        e.progress = 0;
    }
    var prog = eval(e.progress);
    $(spec + " .time").text(format_time(e.time));
    $(spec + " .percent").text(prog.toFixed(0) + "%");
    $(spec + " .message").text(e.message);
    $(spec + " .error").text(e.error);
    $(spec + " .progressbar").css("width", e.progress + "%");

    $(".cometsupport iframe").contents().find("script:not(:last)").remove();
    //$(".cometsupport iframe").contents().find("pre:not(:last)").remove();
}

function comet_update_message(e, msg_class) {
    msg_class = msg_class || 'error';
    if (msg_class == 'finished') // Ignore non-error exit messages
        return;
    var out_class = 'message';
    var header = '';
    var footer = '';
    if (msg_class == 'fatal' || msg_class == 'error') {
        out_class = 'errorbox';
        if (msg_class == 'fatal') {
            header = 'FATAL ERROR: ';
            footer = '<br>Please contact <a href="mailto:support@somat.com">'
                + 'SoMat Customer Support</a> with the above message and the'
                + ' contents of the <a href="/eDAQ/Test/Logbook">eDAQ LogBook'
                + '</a> for assistance.';
        } else {
            header = 'Error: ';
        }
    }
    $("#messages").append(
        '<div class="' + out_class + '">'
        + header + e.message + footer
        + '</div>'
    );
}

function comet_update(e) {
    if (e.type == 'progress') {
        comet_update_progress(e); // KLUDGE use a dispatcher
    } else {
        comet_update_message(e, e.type); // KLUDGE use a dispatcher
    }
}

$(document).ready(function() {
    $(".cometsupport").append('<iframe id="cometiframe" src="__PathAndQuery__&_comet_iframe=1">');
    $("input[name=comet]").attr("value", 1);
    $("#cometiframe").load(function () {
        $("#messages").append(''
            + '<p>\n'
            + '<form method="GET" action="/eDAQ/Hardware">\n'
            + '<input type="submit" name="_Cancel" value="OK">\n'
            + '</form>\n'
            + '</p>\n'
        );
    });
});

</script>
EOT

sub emit_header {
    my ($s) = @_;
    my @updates;
    push @updates, "bundle=$bundle_file";
    push @updates, 'auto_update=' . ($s->param("auto_update") || 0);
    for my $p ($s->get('HTTP_Post_Hash'), $s->get('HTTP_Query_Hash')) {
        for my $k (keys %$p) {
            my $v = $s->param($k);
            push @updates, "$k=1" if $k =~ /^update_/;
            push @updates, "$k=$v" if $k =~ /^fw_/;
        }
    }
    my $q = join('&', @updates);
    $scripts =~ s/__PathAndQuery__/$s->get('HTTP_Path') . "?" . $q/e;
    $s->edaq_header($title, undef, head_extra=>$scripts);
    print "<h3>$title</h3>\n";
}

sub warning_msg {
    my ($flag) = @_;
    my $msg;
    if ($flag eq 'edaqserver-dead') {
        $msg = qq(<p>
The eDAQ's test-processing software has been disabled to allow
firmware updates to take place.  The eDAQ must be restarted to restore
normal operation.
</p>);
    }
    if ($flag eq 'update') {
        $msg = <<"EOT";
<p>
Firmware on the eDAQ system is being updated.
</p>
<p>
<b>Do NOT remove power to the eDAQ or toggle the
front panel or remote power switch during the update!</b>
</p>
EOT
    }
    if ($flag eq 'reboot') {
        $msg = <<"EOT";

<p>The eDAQ system firmware is being updated.</p>
<ul>
<li> The eDAQ will be rebooted.  After the reboot all three LEDs will
     be on.
<li> As the eDAQ is rebooting, the system firmware update will take place.
     The red LED will be blinking at 2 Hz while this is happening.
     <b>Do NOT remove power to the eDAQ or toggle the
     front panel or remote power switch during the update!</b>
     This may take several minutes.  (Diagnostic output will appear on the
     serial port at 115200 bps.)
<li> After the update is complete, the eDAQ will boot normally.
</ul>
<p>
When the eDAQ has successfully rebooted, click the "OK" button to
continue.
</p>
EOT
    }
    return $msg;
}

if ($s->canceled) {
    return;
}

my $mu = SoMat::eDAQ::Hardware::MultiUpdate->new();
my $manifest = $mu->get_manifest($bundle_file);

system "mkdir -p /var/state/web_notifications.d";
if (open(my $notefh, '>', '/var/state/web_notifications.d/edaqserver-dead')) {
    print $notefh warning_msg('edaqserver-dead');
    close $notefh;
}

open(my $dopfh, '>', '/var/state/edaqserver-dead-on-purpose');
close $dopfh;
system "kill -PWR \$(cat /var/run/edaqserver.pid)";
my $timeout = 60;
while ($timeout-- > 0) {
    last unless -f '/var/run/edaqserver.pid';
    die "Timed out waiting for edaqserver to stop" unless $timeout;
    sleep 1;
}

# KLUDGE - Try to shut off ADC_RUN on all boards; if this is on, the
# HLS and BRG cannot read their flash, and will fail detection!  This
# command will fail on some boards, but we don't care.
require SoMat::eDAQ::Hardware::Expmod;
for my $i (0..7) {
    if (my $dev = SoMat::eDAQ::Hardware::Expmod->new("/dev/edaq/expmod/$i")) {
        $dev->send_packet("SET LUN=0 ADC_RUN=OFF");
        $dev->recv_cmd_response;
    }
}

my $hw_rows = [];
my $hw_txt = "$update_dir/hw.txt";
unless (-f $hw_txt) {
    open(my $fh, '>', $hw_txt) or die "Can't write $hw_txt: $!";
    require SoMat::eDAQ::Hardware::Query;
    my $hardware_pm = "$INC[0]/hardware.pm";
    print $fh SoMat::eDAQ::Hardware::Query::gen_hw_tab_db($hardware_pm);
    close($fh);
}
my $hw = SoMat::DB::Tab->new($hw_txt);
$hw->read;
my $i = 1;
while (defined $hw->seek_row($i++)) {
    my $row = $hw->get_row_hash;
    # KLUDGE to deal with boards that won't come up - see Expmod::get_serial
    $row->{serial} =~ s/^X//;
    $row->{version} .= " build $row->{build}"
        if $row->{build} ne '';
    $row->{_signature} = "$row->{tcename}:$row->{serial}";
    my $fws = $row->{_possible_fw} = [];
    for my $mf (@$manifest) {
        my $fw = {};
        if ($row->{_signature} =~ /$mf->{match}/) {
            $fw->{_mf} = $mf;
            $fw->{_new_version} = $mf->{version};
            $fw->{_new_version} .= " build $mf->{build}"
                if defined $mf->{build};
            $fw->{_file_path} = $mf->{file};
            ($fw->{_file_name} = $mf->{file}) =~ s/.*\///;
            $fw->{_checked} = 1 if $row->{version} ne $fw->{_new_version}
                || $row->{connector} =~ /unknown/i
                || (defined $mf->{build} && $mf->{build} eq '0')
                || $s->param("update_$row->{tcename}");
            # Godawful KLUDGE for DIO vbc firmware
            if ($row->{serial} =~ /^X?MSDIO/) {
                my $srcvbc = ($row->{version} =~ /vbc/);
                my $dstvbc = ($fw->{_new_version} =~ /vbc/);
                if (($srcvbc && $dstvbc) || (!$srcvbc && !$dstvbc)) {
                    $row->{_best_fw} = $fw;
                }
            } else {
                $row->{_best_fw} = $fw;
            }
            push @$fws, $fw;
        }
    }
    $row->{_best_fw} ||= $row->{_possible_fw}[0];
    if (defined $row->{_best_fw}) {
        $row->{_checked} = $row->{_best_fw}{_checked};
        push @$hw_rows, $row;
    }
}
undef $hw;

my $node = $s->get('HTTP_Address');

my $comet_iframe = $s->param('_comet_iframe');

if (($s->param('_Commit') || $s->param('auto_update'))
    && $s->param('comet') && !$comet_iframe) {
    emit_header($s);
    print warning_msg('update');
    print qq{<div id="progress"></div>\n};
    print qq{<div id="messages"></div>\n};
    print qq{<div class="cometsupport"></div>\n};
    return $s->edaq_footer();
} elsif ($s->param('auto_update') || $s->param('_Commit') || $comet_iframe) {
    $| = 1;

    system "mkdir -p /var/state/web_notifications.d";
    unlink "/var/state/web_notifications.d/edaqserver-dead";
    if (open(my $notefh, '>', '/var/state/web_notifications.d/updater')) {
        print $notefh warning_msg('update');
        close $notefh;
    }

    if ($comet_iframe) {
        $s->send_header('text/html', 200);
        # Workaround for IE: see http://forums.devx.com/showthread.php?t=163336
        print ' ' x 256;
    } else {
        emit_header($s);
        print warning_msg('update');
    }
    my $updates = [];
    my $time_estimate = 0;
    for my $row (@$hw_rows) {
        if ($s->param('auto_update')) {
            next unless $row->{_checked};
        } else {
            next unless $s->param("update_$row->{tcename}");
            if (my $selected_fw = $s->param("fw_$row->{tcename}")) {
                for my $fw (@{$row->{_possible_fw}}) {
                    if ($fw->{_file_name} eq $selected_fw) {
                        $row->{_best_fw} = $fw;
                        last;
                    }
                }
            }
        }

        if ($row->{_best_fw}) {
            my $mf = $row->{_best_fw}{_mf};
            $row->{firmware_file} = $mf->{file};
            $row->{firmware_type} = $mf->{class};
            $row->{reflash_time} = $mf->{estimated_seconds};
        }
        unless (defined $row->{firmware_file}) {
            if ($comet_iframe) {
                # KLUDGE emit error here?
            } else {
                print "<p>"; # KLUDGE error class
                print "Update of $row->{tcename} $row->{serial} failed: "
                    . "No matching firmware available";
                print "</p>";
            }
            next;
        }
        $time_estimate = (defined($time_estimate) && $row->{reflash_time})
            ? $time_estimate + $row->{reflash_time}
            : undef;
        push @$updates, $row;
    }

    # KLUDGE do something here if no updates

    push(@$updates, {tcename=>'System Reset', firmware_type=>'reset'});

    my $mon = SoMat::Network::ProcessMonitor->new(comet=>$comet_iframe,
        update_interval=>($comet_iframe ? 1 : 5));
    my $dsp = $mon->{dispatcher};
    $mon->{base_time} = time;
    $mon->{elapsed} = -1;

    open(my $ledfile, '>', '/var/state/leds');
    print $ledfile <<EOF;
GREEN_LED_MODE=ON
YELLOW_LED_MODE=ON
RED_LED_MODE=BLINK
RED_LED_ON_TIME=256
RED_LED_OFF_TIME=256
EOF
    system "kill -USR1 \$(cat /var/run/powerd.pid) >/dev/null 2>&1";

    my $total_start = time;

    my $classes = [['exp', 'hlb'], ['vbm'], ['pwr'], ['mpb'], ['reset']];
    for my $pass (0, 1) {
        for my $class (@$classes) {
            for my $type (@$class) {
                for my $row (@$updates) {
                    next unless $row->{firmware_type} eq $type;
                    if ($pass == 0) {
                        $mon->monitor({name=>$row->{tcename}, type=>'progress',
                            buf=>'', time=>0, progress=>0, message=>'Queued'});
                    } elsif ($pass == 1) {
                        my $fn = "update_$type";
                        my $data = {name=>$row->{tcename}, buf=>''};
                        $mon->fork_updater(sub {
                            $mu->$fn($row, $bundle_file, $row->{firmware_file});
                        }, $data);
                    }
                }
            }
            if ($pass == 1) {
                while (@{$dsp->{rd}} || @{$dsp->{wr}}) {
                    $dsp->dispatch;
                }
                last if $mon->fatal_count; # End class loop on fatal errors
            }
        }
    }

    open(my $ledfile, '>', '/var/state/leds');
    print $ledfile <<EOF;
GREEN_LED_MODE=ON
YELLOW_LED_MODE=ON
RED_LED_MODE=ON
EOF
    system "kill -USR1 \$(cat /var/run/powerd.pid) >/dev/null 2>&1";

    if ($comet_iframe) {
        return 1;
    }

    print "<p>";
    printf "Total elapsed time: %d sec", time - $total_start;
    print "</p>\n";

    # KLUDGE conditionalize on rebootage..

    my $f = $s->form(action=>$s->ui_url('Hardware'));
    print $f->begin;
    print $f->submit(name=>'_Cancel', value=>'OK');
    print $f->end;
    return $s->edaq_footer();
} else {
    emit_header($s);

    my $f = $s->form;
    print $f->begin;

    my $t = $s->ruled_shaded_table;
    print $t->begin;
    print $t->row;
    map { print $t->th($_) } (
        'Node', 'Connector', 'Layer Type', 'Serial #', 'Current', 'New',
        'Update'
    );

    for my $row (@$hw_rows) {
        next unless $row->{version};
        next unless $row->{_best_fw};
        next if $row->{tcename} =~ /^GPS_/;

        print $t->row;
        print $t->td($node);
        print $t->td($row->{tcename});
        print $t->td($row->{description});
        print $t->td($row->{serial});
        my $ver = $row->{version};

        print $t->td($ver, 'r');
        print $t->td($row->{_best_fw}{_new_version}, 'r');
        #print $t->td($signature, 'r');

        my $cb_name = "update_$row->{tcename}";
        my $sel_name = "fw_$row->{tcename}";
        my $cell = $f->checkbox(name=>$cb_name, value=>1,
            default=>$row->{_checked});
        if (scalar @{$row->{_possible_fw}} == 1) {
            $cell .= " $row->{_best_fw}{_file_name}";
        } else {
            $cell .= qq(<select name="$sel_name">);
            for my $fw (@{$row->{_possible_fw}}) {
                my $sel = ($fw eq $row->{_best_fw} ? qq( selected="1") : '');
                $cell .= qq(<option value="$fw->{_file_name}"$sel>$fw->{_file_name}</option>);
            }
            $cell .= qq(</select>);
        }
        print $t->td($cell, 'l');
    }
    print $t->end;

    print $f->hidden(name=>'bundle', value=>$bundle_file);
    print $f->hidden(name=>'comet', value=>0);
    print $f->submit(name=>'_Commit', value=>"Update Selected Layer Firmware");
    print $f->submit(name=>'_Cancel', value=>"Cancel"); 
    print $f->end;

    return $s->edaq_footer();
}
